// bagging-ext.cc

#include<strstream.h>
#include"bagging-ext.h"
#include"utils.h"

BaggingExt::BaggingExt() : Bagging(), n_attr(DEFAULT_NUM_ATTR), predicted_classes() {}
										 
/*
	 Constructor.
*/
BaggingExt::BaggingExt(BaseLearnerExt* baselearner, unsigned num_class, unsigned num_attr) :
  Bagging(baselearner, num_class), n_attr(num_attr), predicted_classes() {}
	

/*
	 Overloading += operator.
*/
BaggingExt& BaggingExt::operator+=(const BaggingExt& Bag){
  
	confusion.initialize();
	if (nclass != Bag.nclass)
	  error_exit("Bagging::operator+= : number of classes do not match");
	
	deque<BaseLearner*>::const_iterator it;
	for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) 
	  baselist.push_back((*it)->clone());
  
	deque<string>::const_iterator i;
	for (i = Bag.modellist.begin(); (i != Bag.modellist.end()); i++) 
	  modellist.push_back(*i);
	deque<float>::const_iterator j;
	for (j = Bag.base_learner_train_errors.begin(); (j != Bag.base_learner_train_errors.end()); j++) 
	  base_learner_train_errors.push_back(*j);
	for (j = Bag.base_learner_test_errors.begin(); (j != Bag.base_learner_test_errors.end()); j++) 
	  base_learner_test_errors.push_back(*j);
	for (j = Bag.ensemble_errors.begin(); (j != Bag.ensemble_errors.end()); j++) 
	  ensemble_errors.push_back(*j);
	maxit += Bag.maxit;
  current_it += Bag.current_it;
	
  return *this;
}

/*
	 Aggregates enesembles of base learners.\\
	 It aggregates two bagged ensembles, returning a new bagged ensemble composed through the fusion of two bagged ensembles
*/ 
BaggingExt BaggingExt::operator+(const BaggingExt& Bag){
  
	if (nclass != Bag.nclass)
	  error_exit("Bagging::operator+ : number of classes do not match");
	BaggingExt result((BaseLearnerExt*)base, nclass); // downcast
	
	deque<BaseLearner*>::const_iterator it;
	for (it = baselist.begin(); (it != baselist.end()); it++) 
	  result.baselist.push_back((*it)->clone());
	for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) 
	  result.baselist.push_back((*it)->clone());
  
	deque<string>::const_iterator i;
	for (i = modellist.begin(); (i != modellist.end()); i++) 
	  result.modellist.push_back(*i);
	for (i = Bag.modellist.begin(); (i != Bag.modellist.end()); i++) 
	  result.modellist.push_back(*i);
	deque<float>::const_iterator j;
	for (j = base_learner_train_errors.begin(); (j != base_learner_train_errors.end()); j++) 
	  result.base_learner_train_errors.push_back(*j);
	for (j = Bag.base_learner_train_errors.begin(); (j != Bag.base_learner_train_errors.end()); j++) 
	  result.base_learner_train_errors.push_back(*j);
	for (j = base_learner_test_errors.begin(); (j != base_learner_test_errors.end()); j++) 
	  result.base_learner_test_errors.push_back(*j);	
	for (j = Bag.base_learner_test_errors.begin(); (j != Bag.base_learner_test_errors.end()); j++) 
	  result.base_learner_test_errors.push_back(*j);
	for (j = ensemble_errors.begin(); (j != ensemble_errors.end()); j++) 
	  result.ensemble_errors.push_back(*j);
	for (j = Bag.ensemble_errors.begin(); (j != Bag.ensemble_errors.end()); j++) 
	  result.ensemble_errors.push_back(*j);
	result.maxit = maxit + Bag.maxit;
  result.current_it = current_it + Bag.current_it;
	
  return result;
}



/**
	 Trains the bagged ensemble using a set of resampled data sets, storing models in memory and not in files\\
	 It trains the base learners using data sets previously resampled by bootstrap.
	 The data sets must be named as basenamen, where n varies from startiteration to maxiterations.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models can be stored in files; the name of the files are stored in the list modellist.
	 The models are maintained in RAM memory through the  method train_learner of the base learner.\\
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner).
	 @param basename  base name of the training set. 
	 @param basemodel  base name of the models.
	 @param startiteration start number of iterations(base learner). 
	 @param maxiterations maximun number of iterations(base learners).
	 @param  base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return the number of trained base learners
*/
unsigned BaggingExt::train_base (char* basename,  unsigned startiteration,
                                 unsigned maxiterations, char* base_err_file) {
  
	char trainfile [256];
  int it;
  float err;
	
  ostrstream otrain(trainfile,256,ios::out);
	
  // clear the  error list if startiteration is not consecutive to the previous
	if ((base_learner_train_errors.size() + 1) != startiteration)
	  base_learner_train_errors.clear();
	for (it = startiteration; it <= maxiterations; it++) {
	   // building name of the training files and of the relative models
     otrain.seekp(0,ios::beg);
		 otrain << basename << it << ends;
		 // loading the training set
		 int num_train = wc (trainfile);
		 TrainingSet tset(n_attr, num_train, trainfile);
		 // cloning and training of the base learner
	   BaseLearnerExt* b = (BaseLearnerExt*)base->clone(); // downcast
	   b->set_number(++current_it); 
	   err = b->train_learner(tset);
		 // putting base learner, error and name of the file model into lists
     baselist.push_back(b);
	   base_learner_train_errors.push_back(err);
  }
  if (base_err_file != NULL) {
	  otrain.seekp(0,ios::beg);
		otrain << base_err_file << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(trainfile, base_learner_train_errors);
	}
  return (current_it);
}




/**
	 Trains the bagged ensemble resampling on the fly the data set, storing models in memory and not in files\\
	 It trains the base learners using data sets resampled by bootstrap on the fly.
	 The data sets are resampled using a uniform probability distribution from a given data set.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models are stored only in computer memory and not on mass memory.
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner).
	 @param trainfile  name of the training set to be subsampled. 
	 @param num_base_learners number of the base learners to be trained (oe equivalently number of the training sets to be generated by
	 bootstrap)
	 @param numsamples cardinality of the resampled training sets 
	 param initseed if equal to 0 the seed of the random generator is initialized using the computer clock, otherwise is set to
		                 initseed. Use a number different from 0 if you want to regenerate the same pseudorandom sequence.
	 @param  base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return the effective number of trained base learners
*/
unsigned BaggingExt::train (char* trainfile,  unsigned num_base_learners, unsigned numsamples, 
                            long initseed, char* base_err_file) {
  
  int it;
  float err;
	
	// loading the training set
	int num_train = wc (trainfile);
	TrainingSet tset(n_attr, num_train, trainfile);
	for (it = 1; it <= num_base_learners; it++) {
	   // building the training files by bootstrap
     TrainingSet* tsub = tset.Subsample_with_replacement (numsamples, initseed);		 
		 // cloning and training of the base learner
	   BaseLearnerExt* b = (BaseLearnerExt*)base->clone(); // downcast
	   b->set_number(++current_it); 
	   err = b->train_learner(*tsub);
		 // putting base learner, error and name of the file model into lists
     baselist.push_back(b);
	   base_learner_train_errors.push_back(err);
		 delete tsub;
  }
  if (base_err_file != NULL) {
	  char buf [256];	
    ostrstream oerr(buf,256,ios::out);
	  oerr.seekp(0,ios::beg);
		oerr << base_err_file << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(buf, base_learner_train_errors);
	}
  return (current_it);
}


/*
	 Test the trained base learners stored in files.
*/
void BaggingExt::test_base_learners(TrainingSet& tset, unsigned startiteration, unsigned maxiterations) {
	
	float err;
	unsigned k;		
	
	// clear the  error list if startiteration is not consecutive to the previous
	if ((base_learner_test_errors.size() + 1) != startiteration)
	  base_learner_test_errors.clear();
	// It permits testing using models and results starting at an arbitrary number
	if (baselist.size() <= (maxiterations-startiteration))
	  error_exit("BaggingExt::test_base_learners : number of base learners not sufficient");
	if (baselist.size() == (maxiterations-startiteration+1))
	  k=0;
	else
	  k=startiteration-1;
	// testing of the base learners  
	for (int i=startiteration; i<=maxiterations; i++) {	   
		 BaseLearnerExt * b = (BaseLearnerExt *)baselist[k++]; //downcast
	   err = b->test_learner(tset); //reults stored in the list predicted_classes of b
		 base_learner_test_errors.push_back(err);
	}
}

/*
	 Computes the files of the classes predicted by the ensemble.
*/
void BaggingExt::compute_ensemble_results(TrainingSet& tset, char* predicted, unsigned startiteration, unsigned maxiterations) {
	
  unsigned const SIZE = 256;
	unsigned npattern = 0;
	unsigned k;
	int predicted_class;
	
	unsigned * vclasses = new unsigned[nclass];
	
	// opening file of the classes predicted by the ensemble
	ofstream fpredicted(predicted);	
	if (fpredicted == NULL)
	  error_exit("BaggingExt::compute_ensemble_results : Can't open file ", predicted);
		
	// It permits testing using models and results starting at an arbitrary number
	if (baselist.size() <= (maxiterations-startiteration))
	  error_exit("BaggingExt::compute_ensemble_results : number of base learners not sufficient");
	if (baselist.size() == (maxiterations-startiteration+1))
	  k=0;
	else
	  k=startiteration-1;
	
	// Reading the classes computed by the base learners, one pattern at a time and storing
	// them into the vector vclasses and computes the predicted class by majority voting
	npattern = tset.Ntrain();
	unsigned pclass;
	unsigned z;
	for (int j=0; j<npattern; j++) {
	  for (int t = 0; t<nclass; t++)
		  vclasses[t] = 0;
    z = k; //recall that startiteration can be different from 1
		for (int i=startiteration; i<=maxiterations; i++) {
		  BaseLearnerExt* b = (BaseLearnerExt*)baselist[z++]; //downcast
			pclass = b->predicted_classes[j];
			vclasses[--pclass]++;
		}
		predicted_class = majority_voting(vclasses);
		predicted_classes.push_back(predicted_class);
		fpredicted << predicted_class << endl;
	}
	if (fpredicted.bad())
	  error_exit("Bagging::compute_ensemble_results : Error writing classes predicted by the base learners in file ", predicted);
		
	// cleaning up
	delete [] vclasses;
}

float BaggingExt::calc_error_and_confusion(TrainingSet& tset, char* base_err_file) {	
	 	 
	int predicted_class;
	int trueclass;
	unsigned npattern;
	float err;
	unsigned num_err = 0;
	
	confusion.initialize();	
	npattern = tset.Ntrain();
	vect& target = tset.Read_target();
	for (unsigned i = 0 ; i<npattern; i++) {
	  trueclass = (unsigned)target[i];
		predicted_class = predicted_classes[i];
		confusion(predicted_class-1, trueclass-1) += 1;
		if (predicted_class != trueclass)
		  num_err++;
	}	
	err = (float)num_err/(float)npattern;
	if (base_err_file != NULL) {
	  char buf [256];	
    ostrstream oerr(buf,256,ios::out);
	  oerr.seekp(0,ios::beg);
		oerr << base_err_file << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(buf, base_learner_test_errors);
	}
	return err;
}	

/*
	 Combines the bagged ensemble of trained base learners stored in memory.
*/														
float BaggingExt::combine (char* testfile, char* predicted, unsigned startiteration,
                           unsigned maxiterations, char* base_err_file) {
  
	float err;
	// loading the test set
	int num_train = wc (testfile);
	TrainingSet tset(n_attr, num_train, testfile);			
	// Tests the trained base learners stored in files
	test_base_learners(tset, startiteration, maxiterations);
	// Computes the files of the classes predicted by the ensemble
	compute_ensemble_results(tset, predicted,  startiteration,  maxiterations);
	// Computes the error and the confusion matrix of the ensemble
	err = calc_error_and_confusion(tset, base_err_file);
  return err;
}	

/*
	 Computes the ensemble errors at each iteration.
*/
void BaggingExt::compute_ensemble_results_step(TrainingSet& tset, char* predicted, unsigned startiteration, unsigned maxiterations){
  
	unsigned npattern = 0;
	unsigned i,j,k;
	int predicted_class;
	int num_err = 0;
	float ens_error;
		
	// It permits testing using models and results starting at an arbitrary number
	if (baselist.size() <= (maxiterations-startiteration))
	  error_exit("BaggingExt::compute_ensemble_results : number of base learners not sufficient");
	if (baselist.size() == (maxiterations-startiteration+1))
	  k=0;
	else
	  k=startiteration-1;
	
	// Reading the targets and intializing the collecting the predictions of the base learners
	npattern = tset.Ntrain();
	vect& target = tset.Read_target(); // vector of the true class labels
	unsigned vpredicted[npattern]; // vector of the class labels predicted by the ensemble
	unsigned pc[npattern][nclass]; // matrix collecting the predictions of the base learners
	for (i=0; i<npattern; i++)
	  for (j=0; j<nclass; j++)  
		  pc[i][j] = 0;
	unsigned pclass;
	unsigned z = k; //recall that startiteration can be different from 1
	
	// calculates the erro rate at each iteration
	for (int i=startiteration; i<=maxiterations; i++) {
	  num_err = 0;	
		BaseLearnerExt* b = (BaseLearnerExt*)baselist[z++]; //downcast		  
		for (j=0; j<npattern; j++)  {
		  pclass = b->predicted_classes[j];
			pc[j][pclass-1]++;
			vpredicted[j] = majority_voting(pc[j]);
			if (vpredicted[j] !=((unsigned)target[j]))
			  num_err++;
		}
		ens_error = (float)num_err/(float)npattern;
		ensemble_errors.push_back(ens_error);
	}
	
	// saving the file of the classes predicted by the ensemble
	ofstream fpredicted(predicted);	
	if (fpredicted == NULL)
	  error_exit("BaggingExt::compute_ensemble_results : Can't open file ", predicted);
	for (j=0; j<npattern; j++) {
	  fpredicted << vpredicted[j] << endl;
		predicted_classes.push_back(vpredicted[j]);
	}
	if (fpredicted.bad())
	    error_exit("BaggingExt::compute_ensemble_results_step : Error writing file ", predicted);
	fpredicted.close();
	
	// saving the ensemble errors at each iteration
	save_ensemble_errors(predicted);
}


/*
	 Combines the bagged ensemble of trained base learners stored in memory and storing the results at each iteration.
*/														
float BaggingExt::combine_step (char* testfile, char* predicted, unsigned startiteration,
                                unsigned maxiterations, char* base_err_file) {
  
	float err;
	// loading the test set
	int num_train = wc (testfile);
	TrainingSet tset(n_attr, num_train, testfile);	
	// Tests the trained base learners stored in files
	test_base_learners(tset, startiteration, maxiterations);
	// Computes the files of the classes predicted by the ensemble
	compute_ensemble_results_step(tset, predicted,  startiteration,  maxiterations);	
	// Computes the error and the confusion matrix of the ensemble
	err = calc_error_and_confusion(tset, base_err_file);
  return err;
}	

/*
   Saves the learned models of the base learners.
*/
void BaggingExt::save_models (char* basemodel, unsigned start, unsigned end){
  
	char modelfile [256];
	unsigned i,k;
	
	ostrstream omodel(modelfile,256,ios::out);
  if (baselist.size() <= (end-start))
	  error_exit("BaggingExt::save_models : Trying to save more base learners than than those stored in memory");		
	if (baselist.size() == (end-start+1))
	  k = 0;
	else
	  k = start -1;
  
	for (i = start; i <= end; i++) { 
	  omodel << basemodel << i << ends;
		string s(modelfile);
		modellist.push_back(s);
		BaseLearnerExt* b = (BaseLearnerExt*)baselist[k++]; //downcast
	  b->save_model(modelfile);
	}	
}



