/* ******************************************************************************
NEURObjects 2.0
bagging-ext.h May 2002
Modified  May 2002
****************************************************************************** */
#ifndef BAGGINGEXT_H
#define BAGGINGEXT_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include "matrix.h"
#include "tset.h"
#include "baselearner.h"
#include "bagging.h"

using namespace std;


/**
   Class building  ensembles of bagged base learners.\\
	 This class implements the general Breiman's learning algorithm of bagging, selecting a class by majority voting
	 among the trained bagged base learners.
	 The base learners must belong to a derived class of BaseLearner. A pointer to a specific base learner (base) is passed to the
	 constructor: each new added base learner will be a copy  of base. 
	 Learned models and bootstrapped data sets are stored in files. Bootstrapped data sets are stored in files consecutively
	 numbered (the number are the last character of the common name of the data sets (that is name1, name2, ... , namen).
	 This class permits to easily merge different bagged ensemble, and it allows to use different base learners 
	 (as long as the base learners are derived classes of the abstract virtual base class Baselearner) 
*/ 
class BaggingExt : public Bagging {

protected:

/// number of the attributes of the training set
unsigned n_attr;

/**
	 Test the trained base learners stored in computer memory.\\
	 Testing of the base learners on the test set tset, using models are stored in computer memory. 
	 Predicted classes of each base learner are stored in lists owned by the base learner.
	 @param testfile  testing set. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 	
*/
void test_base_learners(TrainingSet& tset, unsigned startiteration, unsigned maxiterations);

/**
	 Computes the classes predicted by the ensemble.\\
	 It performs a test of the bagged ensembled on an assignated test set, after that the base learners have been tested on the test set.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors.
	 The results computed by the base learners are stored in the list predicted_classes.
	 The results computed by each base learners (list of the predicted classes)  are stored  in  the list predicted_classes
	 owned by the base learners of the class BaseLearnerExt.
	 The predicted class of the ensemble is assigned through majority voting
	 @param tset  test set.
	 @param predicted file storing the predicted classes. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 	
*/
void compute_ensemble_results(TrainingSet& tset, char* predicted, unsigned startiteration, unsigned maxiterations);

/**
	 Computes the error and the confusion matrix of the ensemble.\\
	 @param tset  testing set.
	 @param  base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return ensemble error rate
*/
float calc_error_and_confusion(TrainingSet& tset, char* base_err_file);

/**
	 Computes the ensemble errors at each iteration.\\
	 Computes the ensemble errors at each iteration and stores in the file predicted the classes predicted by the ensemble at the last
	 iteration.
	 At each iteration the predicted class of the ensemble is assigned through majority voting. The errors of the ensemble at each
	 iteration is stored in the list ensemble_errors and in the file predicted.ens_errors
	 The  classes predicted by  the ensemble at the last iteration are stored in the file predicted.
	 @param tset  testing set.
	 @param predicted file storing the predicted classes. 
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 	
*/
void compute_ensemble_results_step(TrainingSet& tset, char* predicted, unsigned startiteration, unsigned maxiterations);

public:

// members

/// List of the calsses predicted by the ensemble
deque<unsigned> predicted_classes;

// methods ///////////////////////////

/**
	 Dummy constructor.	
	 It does nothing 
*/
  BaggingExt();

/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param num_class number of classes;
	 @param num_attr number of attributes;
*/
BaggingExt(BaseLearnerExt* baselearner, unsigned num_class = DEFAULT_NUM_CLASS, unsigned num_attr = DEFAULT_NUM_ATTR);


/**
	 Destructor.
*/
 virtual ~BaggingExt(){}

/**
	 Aggregates enesembles of base learners.\\
	 It aggregates two bagged ensembles, returning a new bagged ensemble composed through the fusion of two bagged ensembles
*/ 
BaggingExt operator+(const BaggingExt& Bag);

/**
	 Overloading += operator.\\
	 It aggregates two bagged ensembles, adding the base learners and the modellist of the bagged ensemble on the right to
	 the bagged ensemble on the left.
*/
BaggingExt& operator+=(const BaggingExt& Bag);

/**
	 Trains the bagged ensemble using a set of resampled data sets, storing models in memory\\
	 It trains the base learners using data sets previously resampled by bootstrap.
	 The data sets must be named as basenamen, where n varies from startiteration to maxiterations.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models can be stored in files; the name of the files are stored in the list modellist.
	 The models are maintained in RAM memory through the  method train_learner of the base learner.\\
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner).
	 @param basename  base name of the training set. 
	 @param basemodel  base name of the models.
	 @param startiteration start number of iterations(base learner). 
	 @param maxiterations maximun number of iterations(base learners).
	 @param  base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return the number of trained base learners
*/
virtual unsigned train_base (char* basename,  unsigned startiteration = START_ITERATIONS,
                             unsigned maxiterations = MAX_ITERATIONS, char* base_err_file = NULL);											
/**
	 Trains the bagged ensemble resampling on the fly the data set, storing models in memory and not in files\\
	 It trains the base learners using data sets resampled by bootstrap on the fly.
	 The data sets are resampled using a uniform probability distribution from a given data set.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models are stored only in computer memory and not on mass memory.
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner).
	 @param trainfile  name of the training set to be subsampled. 
	 @param num_base_learners number of the base learners to be trained (oe equivalently number of the training sets to be generated by
	 bootstrap)
	 @param numsamples cardinality of the resampled training sets 
	 param initseed if equal to 0 the seed of the random generator is initialized using the computer clock, otherwise is set to
		                 initseed. Use a number different from 0 if you want to regenerate the same pseudorandom sequence.
	 @param  base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return the effective number of trained base learners
*/
unsigned train (char* trainfile,  unsigned num_base_learners, unsigned numsamples, 
                            long initseed = 0, char* base_err_file = NULL);

/**
	 Combines the bagged ensemble of trained base learners stored in memory.\\
	 It performs a test of the bagged ensembled on an assignated test set (testfile). 
	 The format of the testfile must be in the format supported by the class TrainingSet.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors.
	 The trained base learners are stored in memory in the list baselist.	  
	 The results computed by each base learners (list of the predicted classes)  are stored  in  the list predicted_classes
	 owned by the base learners of the class BaseLearnerExt.
	 The predicted class of the ensemble is assigned through majority voting.\\
	 Output: error rate, confusion matrix and file of the classes predicted by the ensemble.
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return error rate
*/														
float combine (char* testfile, char* predicted, unsigned startiteration = START_ITERATIONS,
                             unsigned maxiterations = MAX_ITERATIONS, char* base_err_file = NULL);
														 
/**
	 Combines the bagged ensemble of trained base learners stored in memory and storing the results at each iteration\\
	 It performs a test of the bagged ensembled on an assignated test set (testfile). 
	 The format of the testfile must be in the format supported by the class TrainingSet.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors and the ensemble error rate
	 in the list ensemble_errors and in the file predicted.ens_errors.
	 The trained base learners are stored in memory in the list baselist.	  
	 The results computed by each base learners (list of the predicted classes)  are stored  in  the list predicted_classes
	 owned by the base learners of the class BaseLearnerExt.
	 The predicted class of the ensemble is assigned through majority voting.\\
	 Output: error rate, confusion matrix, file of the classes predicted by the ensemble, file of the ensemble error at each iteration
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 base_err_file if not equal to NULL the errors of the base learners are stored into the file base_err_file.blerr
	 @return error rate
*/														
float combine_step (char* testfile, char* predicted, unsigned startiteration = START_ITERATIONS,
                    unsigned maxiterations = MAX_ITERATIONS, char* base_err_file = NULL);

/**
   Saves the learned models of the base learners.
   The trained models are stored in files; the name of the files are stored in the list modellist.
	 The modelllist contain the name of the models (whose name is basemodeln, with n going from start to
	 end). The nmber of the stored models is end-start+1.
	 @param basemodel base name of  models to be stored
	 @param starting number of the model
	 @param ending number of the models
*/
void save_models (char* basemodel, unsigned start, unsigned end);

};

#endif
