/* ******************************************************************************
NEURObjects 2.0
bagging.h January 2002
Modified  May 2002
Giorgio Valentini
****************************************************************************** */
#ifndef BAGGING_H
#define BAGGING_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include "matrix.h"
#include "tset.h"
#include "baselearner.h"

using namespace std;

/**
   Class for bagging ensembles of base learners.\\
	 This class implements the general Breiman's learning algorithm of bagging, selecting a class by voting
	 among the trained base learners.\\
	 This class uses homogenous base learners, that is base learner of the same type. The base
	 learner must be a derived class of BaseLearner. A pointer to a specific base learner (base) is passed to the
	 constructor: each new added base learner will be a copy  of base.
*/ 
class Bagging {

protected:
/// Base learner model. Each new added base learner is equal to base.
BaseLearner* base;

/// Default starting number of iterations
static const unsigned START_ITERATIONS = 0;

/// Default maximun number of iterations
static const unsigned MAX_ITERATIONS = 100;

/// Default  number of classes
static const unsigned DEFAULT_NUM_CLASS = 2;

/// Default  number of attributes
static const unsigned DEFAULT_NUM_ATTR = 2;

static const char * const SUFFIX_BASE_LEARNER_ERRORS = ".blerr";

/// Confusion matrix
matrix confusion;

/// number of classes
unsigned nclass;

/// number of the attributes of the training set
// unsigned n_attr;

public:

// members

/// List of the base learners
deque<BaseLearner*> baselist;

/**  List of the model files.\\
     It stores the names of the files of the trained base learners.
*/
deque<string> modellist;

/**  List of the training errors of the base learners.\\
     It stores the training error rate of each base learner composing the ensemble.
*/
deque<float> base_learner_train_errors;

/**  List of the test errors of the base learners.\\
     It stores the test error rate of each base learner composing the ensemble.
*/
deque<float> base_learner_test_errors;

/**  List of the error rates of the ensemble at each iteration .\\
     It stores the error rate of the ensemble at each iteration.
*/
deque<float> ensemble_errors;

/// Maximun number of iterations(base learners).
unsigned maxit;

/**  Current number of iterations(base learners). \\
The numbering starts from 1, and  the number of the base learners is  current_it;
*/
unsigned current_it;

/// Training set
TrainingSet* trainset;

/// Test set
TrainingSet* testset;



// methods

/**
	 Dummy constructor.	
	 It does nothing 
*/
  Bagging();

/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param nclass number of classes;
*/
Bagging(BaseLearner* baselearner, unsigned nclass = 2);

/**
	 Copy Constructor.
*/
  Bagging(const Bagging&);


/**
	 Assignment operator.
*/
  Bagging& operator=(const Bagging&){}

/**
	 Destructor.
*/
 virtual ~Bagging();





/**
	Prints info about the bagged ensemble
	@param O output stream
*/ 
void print_info(ostream& O){}

/**
	Prints info about the bagged ensemble to cout
*/ 
void print_info() { print_info(cout); }

/**
	Prints the confusion matrix to an output stream
*/ 
void print_conf_matrix(ostream& os = cout);

/**
	Gets the pointer to the base learner of the ensemble
	@return pointer to a base learner
*/ 
BaseLearner* get_base_learner(void){ return base;}

/**
	Sets the base learner of the ensemble
	@param b base learner
*/ 
void set_base_learner(BaseLearner* b){ base = b;}

/**
	Saves into a file the error rate of the ensemble at each iteration
	@param filename file of the error rates of the ensemble
*/ 
void save_ensemble_errors(char* filename){}

/**
	Saves into a file the error rate of each base learner of the ensemble
	@param filename file of the error rates of the base learners
	@param b list of the errors
*/ 
void save_base_learner_errors(char* filename, deque<float>& b);
											 

/**
	Computes by majority voting the class
	Using a vector of the classes voted by the base learners it calculates the class predicted by the ensemble by majority voting.
	If two or more classes achieve the same maximal number of votes a random guess is performed.
	@param vclasses vector of the classes voted by the base learners
*/ 
unsigned majority_voting(unsigned* vclasses);


/**
	 Trains the bagged ensemble using a set of resampled data sets.\\
	 It trains the base learners using data sets previously resampled by bootstrap.
	 The data sets must be named as basename.tn, where n varies from startiteration to maxiterations.
	 At each iteration it stores  error training rate of each base learner in the list base_learner_train_errors. 
	 The trained models can be stored in files; the name of the files are stored in the list modellist.
	 The modelllist contain the name of the models (whose name is basemodel-mn, with n going from startiteration to
	 maxiterations). The models are stored through the method train_learner of the base learner.\\
	 Note that you can call more times this method, eventually using different base learners (calling set_base_learner),
	 but hte data sets must be numbered consecutively starting from 1.\\
	 @param basename  base name of the training set. 
	 @param basemodel  base name of the models.
	 @param startiteration start number of iterations(base learner). 
	 @param maxiterations maximun number of iterations(base learners).
	 @param  save_err if true the errors of the base learners are stored into the file basemodel.blerr
	 @return the number of trained base learners
*/
virtual unsigned train_base_model (char* basename, char* basemodel, unsigned startiteration = START_ITERATIONS,
                                   unsigned maxiterations = MAX_ITERATIONS, bool save_err = false);


/**
	 Combines the bagged ensemble of trained base learners using model files.\\
	 It performs a test of the bagged ensembled on an assignated test set (testfile). 
	 The format of the testfile is free, but it must be understandable to the base learner.
	 A second test file named testfile.labels must be provided. It contains the list of the labels
	 of the classes of the test file in the same format of the predicted file of the class BaseLearner.
	 At each iteration it stores  error test rate of each base learner in the list base_learner_test_errors.
	 The names of the files of the trained base learners are stored in the list modellist.
	 Each trained base learner is loaded from the files stored in model list and it is tested on all 
	 the test set. 
	 The results (list of the predicted classes) are stored  in temporary files from 
	 basres1 to basresn, where n id the number of the base learners.
	 Then all the files are opened and each pattern is read one at a time from all the files and 
	 the predicted class of the enesemble is assigned through majority voting.\\
	 Output: error rate, confusion matrix and file of the classes predicted by the ensemble.
	 @param testfile  name of the testing set. 
	 @param predicted file storing the predicted classes.
	 @param startiteration number of the first base learner to be used. 
	 @param maxiterations number of the last base learner to be used. 
	 @param  save_err if true the errors of the base learners are stored into the file predicted.blerr
	 @return error rate
	 @see BaseLearner
*/
virtual float combine_models (char* testfile, char* predicted, unsigned startiteration = START_ITERATIONS,
                              unsigned maxiterations = MAX_ITERATIONS, bool save_err = false);

};





#endif
