// bagging.cc

#include<strstream.h>
#include"bagging.h"
#include"utils.h"

Bagging::Bagging() : confusion(), nclass(0), 
                     baselist(), modellist(), base_learner_train_errors(), base_learner_test_errors(), ensemble_errors(),
										 maxit(MAX_ITERATIONS), current_it(0)
{ base = NULL; trainset = NULL; testset = NULL;}
										 
/**
	 Constructor.
	 @param baselearner base learner to be used by the bagging algorithm
	 @param nclass number of classes;
*/
Bagging::Bagging(BaseLearner* baselearner, unsigned num_class) :
  confusion(num_class,num_class,0.0), nclass(num_class), baselist(), modellist(), 
	base_learner_train_errors(), base_learner_test_errors(), ensemble_errors(), maxit(MAX_ITERATIONS), current_it(0)
{
  base = baselearner;
  trainset = NULL; 
	testset = NULL;
}

/**
	 Copy Constructor.
*/
Bagging::Bagging(const Bagging& Bag) {
 // n_attr = Bag.n_attr;
 nclass = Bag.nclass;
 confusion = Bag.confusion;
 
 // copy of the base learners of Bag into this
 deque<BaseLearner*>::const_iterator it;
 for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) {
	 baselist.push_back((*it)->clone());
 }  
 
 modellist = Bag.modellist;
 base_learner_train_errors = Bag.base_learner_train_errors;
 base_learner_test_errors = Bag.base_learner_test_errors;
 ensemble_errors = Bag.ensemble_errors;
 maxit = Bag.maxit;
 current_it = Bag.current_it;
 base = Bag.base;
 trainset = Bag.trainset;
 testset = Bag.testset; 
}

/**
	 Assignment operator.
*/
  Bagging& Bagging::operator=(const Bagging& Bag) {
	
	// destruction of the base learners 
 deque<BaseLearner*>::const_iterator it;
 for (it = baselist.begin(); (it != baselist.end()); it++) 
	 delete *it;
 
 baselist.clear();
 for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) 
	 baselist.push_back((*it)->clone());
 
 nclass = Bag.nclass;
 confusion = Bag.confusion;
 
 modellist.clear();
 base_learner_train_errors.clear();
 base_learner_test_errors.clear();
 ensemble_errors.clear();
 modellist = Bag.modellist;
 base_learner_train_errors = Bag.base_learner_train_errors;
 base_learner_test_errors = Bag.base_learner_test_errors;
 ensemble_errors = Bag.ensemble_errors;
 maxit = Bag.maxit;
 current_it = Bag.current_it;
 base = Bag.base; 
 if (trainset != NULL) {
	  delete trainset;
		trainset = NULL;
  }	
	if (testset != NULL) {
	  delete testset;
		testset = NULL;
  }
  trainset = Bag.trainset;
  testset = Bag.testset; 

	return *this;
}

/* Returns a specific base learner from the bagged ensembles */
BaseLearner* Bagging::operator[](int i) {
  if ((i<0) || ((i+1)>baselist.size()))
	   error_exit("Bagging::operator[] : out of range");
	return (baselist[i]);
} 

/*
	 Overloading += operator.
*/
Bagging& Bagging::operator+=(const Bagging& Bag){
  
	confusion.initialize();
	if (nclass != Bag.nclass)
	  error_exit("Bagging::operator+= : number of classes do not match");
	
	deque<BaseLearner*>::const_iterator it;
	for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) 
	  baselist.push_back((*it)->clone());
  
	deque<string>::const_iterator i;
	for (i = Bag.modellist.begin(); (i != Bag.modellist.end()); i++) 
	  modellist.push_back(*i);
	deque<float>::const_iterator j;
	for (j = Bag.base_learner_train_errors.begin(); (j != Bag.base_learner_train_errors.end()); j++) 
	  base_learner_train_errors.push_back(*j);
	for (j = Bag.base_learner_test_errors.begin(); (j != Bag.base_learner_test_errors.end()); j++) 
	  base_learner_test_errors.push_back(*j);
	for (j = Bag.ensemble_errors.begin(); (j != Bag.ensemble_errors.end()); j++) 
	  ensemble_errors.push_back(*j);
	maxit += Bag.maxit;
  current_it += Bag.current_it;
	
  return *this;
}

/*
	 Aggregates enesembles of base learners.\\
	 It aggregates two bagged ensembles, returning a new bagged ensemble composed through the fusion of two bagged ensembles
*/ 
Bagging Bagging::operator+(const Bagging& Bag){
  
	if (nclass != Bag.nclass)
	  error_exit("Bagging::operator+ : number of classes do not match");
	Bagging result(base, nclass);
	
	deque<BaseLearner*>::const_iterator it;
	for (it = baselist.begin(); (it != baselist.end()); it++) 
	  result.baselist.push_back((*it)->clone());
	for (it = Bag.baselist.begin(); (it != Bag.baselist.end()); it++) 
	  result.baselist.push_back((*it)->clone());
  
	deque<string>::const_iterator i;
	for (i = modellist.begin(); (i != modellist.end()); i++) 
	  result.modellist.push_back(*i);
	for (i = Bag.modellist.begin(); (i != Bag.modellist.end()); i++) 
	  result.modellist.push_back(*i);
	deque<float>::const_iterator j;
	for (j = base_learner_train_errors.begin(); (j != base_learner_train_errors.end()); j++) 
	  result.base_learner_train_errors.push_back(*j);
	for (j = Bag.base_learner_train_errors.begin(); (j != Bag.base_learner_train_errors.end()); j++) 
	  result.base_learner_train_errors.push_back(*j);
	for (j = base_learner_test_errors.begin(); (j != base_learner_test_errors.end()); j++) 
	  result.base_learner_test_errors.push_back(*j);	
	for (j = Bag.base_learner_test_errors.begin(); (j != Bag.base_learner_test_errors.end()); j++) 
	  result.base_learner_test_errors.push_back(*j);
	for (j = ensemble_errors.begin(); (j != ensemble_errors.end()); j++) 
	  result.ensemble_errors.push_back(*j);
	for (j = Bag.ensemble_errors.begin(); (j != Bag.ensemble_errors.end()); j++) 
	  result.ensemble_errors.push_back(*j);
	result.maxit = maxit + Bag.maxit;
  result.current_it = current_it + Bag.current_it;
	
  return result;
}

/* Destructor.*/
Bagging::~Bagging() {
  deque<BaseLearner*>::iterator it;
  for (it = baselist.begin(); (it != baselist.end()); it++) 
	  delete *it;
		
	if (trainset != NULL) {
	  delete trainset;
		trainset = NULL;
  }
	
	if (testset != NULL) {
	  delete testset;
		testset = NULL;
  }
}


/* Trains the bagged ensemble using a set of resampled data sets.*/
unsigned Bagging::train_base_model (char* basename, char* basemodel, unsigned startiteration,
                                   unsigned maxiterations, bool save_err, bool b_printout) {
  
	char trainfile [256];
	char modelfile [256];
  int it;
  float err;
	
  ostrstream otrain(trainfile,256,ios::out);
	ostrstream omodel(modelfile,256,ios::out);
  
	if (b_printout)
	  cout << "Training " << flush;
  // clear the  error list if startiteration is not consecutive to the previous
	if ((base_learner_train_errors.size() + 1) != startiteration)
	  base_learner_train_errors.clear();
	for (it = startiteration; it <= maxiterations; it++) {
	   // building name of the trainig files and of the relative models
     otrain.seekp(0,ios::beg);
		 omodel.seekp(0,ios::beg);		 
		 otrain << basename << it << ends;
		 omodel << basemodel << "-m" << it << ends;
		 // cloning and training of the base learner
	   BaseLearner* b = base->clone();
	   b->set_number(++current_it); 
	   err = b->train_learner(trainfile,modelfile);
		 // putting base learner, error and name of the file model into lists
     baselist.push_back(b);
	   base_learner_train_errors.push_back(err);
		 string s(modelfile);
		 modellist.push_back(s);
		 if (b_printout)
	     cout << "." << flush;
  }
  if (save_err) {
	  omodel.seekp(0,ios::beg);
		omodel << basemodel << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(modelfile, base_learner_train_errors);
	}
	if (b_printout)
	     cout << endl;
  return (current_it);
}

/*
	 Test the trained base learners stored in files.
*/
void Bagging::test_base_learners(char* testfile, unsigned startiteration, 
                                unsigned maxiterations, char* dir_res) {
	
	unsigned const SIZE = 256;
	char model[SIZE];
	char results[SIZE];
	float err;
	unsigned k;		
	
	ostrstream omodel(model,SIZE,ios::out);
	ostrstream oresults(results,SIZE,ios::out);
	// clear the  error list if startieration is not consecutive to the previous
	if ((base_learner_test_errors.size() + 1) != startiteration)
	  base_learner_test_errors.clear();
	// It permits testing using models and results starting at an arbitrary number
	if (baselist.size() <= (maxiterations-startiteration))
	  error_exit("Bagging::test_base_learners : number of base learners not sufficient");
	if (baselist.size() == (maxiterations-startiteration+1))
	  k=0;
	else
	  k=startiteration-1;
	// testing of the base learners  
	for (int i=startiteration; i<=maxiterations; i++) {	   
		 omodel.seekp(0,ios::beg);	
		 oresults.seekp(0,ios::beg);
		 //if ((i>modellist.size()) || (i<1))
		 //  error_exit("Bagging::test_base_learners : modellist out of range");	 		 
		 omodel << modellist[k] << ends;
     if (!strcmp(dir_res,""))
       oresults  << BASRES << i << ends;
     else
		   oresults << dir_res << "/" << BASRES << i << ends;
		 //if ((i>baselist.size()) || (i<1))
		 //  error_exit("Bagging::test_base_learners : baselist out of range");	
	   err = (baselist[k++])->test_learner(testfile, model, results);
		 base_learner_test_errors.push_back(err);
	}
}

/*
	 Computes the files of the classes predicted by the ensemble.
*/
void Bagging::compute_ensemble_results(char* testfile, char* predicted, unsigned startiteration, unsigned maxiterations) {
	
  unsigned const SIZE = 256;
	char results[SIZE];
	char testlabels[SIZE];
	unsigned npattern = 0;
	unsigned k;
	int predicted_class;
	
	unsigned * vclasses = new unsigned[nclass];
	ifstream * ifs = new ifstream[maxiterations-startiteration+1];	
	ostrstream oresults(results,SIZE,ios::out);
	
	// Combining the results using majority voting and the files basres
	// opening file of the classes predicted by the ensemble
	ofstream fpredicted(predicted);	
	if (fpredicted == NULL)
	  error_exit("Bagging::compute_ensemble_results : Can't open file ", predicted);
	// opening files of the predicted classes computed by the base learners
	k = 0; // recall that startiteration can be not equal to 0
	for (int i=startiteration; i<=maxiterations; i++) {
	  oresults.seekp(0,ios::beg);
	  oresults << BASRES << i << ends;
	  ifs[k++].open(results);
		if (ifs[k-1] == NULL)
		  error_exit("Bagging::compute_ensemble_results : Can't open file ", results);
	}
	
	// Reading the classes computed by the base learners, one pattern at a time and storing
	// them into the vector vclasses and computes the predicted class by majority voting
	ostrstream otestlabels(testlabels,SIZE,ios::out);
	otestlabels << testfile << ".labels" << ends;
	npattern = wc(testlabels);
	unsigned pclass;
	for (int j=0; j<npattern; j++) {
	  for (int t = 0; t<nclass; t++)
		  vclasses[t] = 0;
		k = 0; // recall that startiteration can be not equal to 0
		for (int i=startiteration; i<=maxiterations; i++) {
		  ifs[k++] >> pclass;
			vclasses[--pclass]++;
		}
		predicted_class = majority_voting(vclasses);
		fpredicted << predicted_class << endl;
	}
	if (fpredicted.bad())
	  error_exit("Bagging::compute_ensemble_results : Error writing classes predicted by the base learners in file ", predicted);
	
	// closing files of the predicted classes computed by the base learners
	k = 0; // recall that startiteration can be not equal to 0
	for (int i=startiteration; i<=maxiterations; i++) {
	    if (ifs[k].bad())
			  error_exit("Bagging::compute_ensemble_results : Error reading classes predicted by the base learners ");
		  ifs[k++].close();
	}
		
	// cleaning up
	delete [] vclasses;
	delete[] ifs;
}


/**
	 Computes the ensemble errors at each iteration.
*/
void Bagging::compute_ensemble_results_step(char* testfile, char* predicted, unsigned startiteration, unsigned maxiterations,
                                            bool save_pred_at_each_iteration) {
	
  unsigned const SIZE = 256;
	char results[SIZE];
	char testlabels[SIZE];
	unsigned npattern = 0;
	unsigned i,j,k;
	int num_err = 0;
	float ens_error;
	
	// Reading the file testfile.labels and storing the labels of the classes
	// Construction of the vector of the class labels predicted by the ensemble
	// and of the matrix (npattern*nclass) storing the predictions of the base learners
	ostrstream otestlabels(testlabels,SIZE,ios::out);
	otestlabels << testfile << ".labels" << ends;
	npattern = wc(testlabels);
	unsigned target[npattern]; // vector of the true class labels
	unsigned vpredicted[npattern]; // vector of the class labels predicted by the ensemble
	unsigned pc[npattern][nclass]; // matrix collectiong the predictions of the base learners
	for (i=0; i<npattern; i++)
	  for (j=0; j<nclass; j++)  
		  pc[i][j] = 0;
	ifstream ftest(testlabels);	
	if (ftest == NULL)
	  error_exit("Bagging::compute_ensemble_results_step : Can't open file ", testlabels);
	for (unsigned i=0; i<npattern; i++)
	  ftest >> target[i];
	if (ftest.bad())
	  error_exit("Bagging::compute_ensemble_results_step : Error reading file ", testlabels);
	ftest.close();
	
	// Computes the error of the ensemble at each iteration and stores it in the list ensemble_errors
	ifstream ifs; //stream of the classes predicted by the base learners	
	ostrstream oresults(results,SIZE,ios::out);	// stream for building name othe files of the classes predicted by the base learners
	unsigned pclass;
	for (int i=startiteration; i<=maxiterations; i++) {
	  num_err = 0;
	  oresults.seekp(0,ios::beg);
	  oresults << BASRES << i << ends;
		ifs.open(results);
	  if (ifs == NULL)
		  error_exit("Bagging::compute_ensemble_results_step : Can't open file ", results);
		for (j=0; j<npattern; j++)  {
		  ifs >> pclass;
			pc[j][pclass-1]++;
			vpredicted[j] = majority_voting(pc[j]);
			if (vpredicted[j] != target[j])
			  num_err++;
		}
	  if (ifs.bad())
	    error_exit("Bagging::compute_ensemble_results_step : Error reading file ", results);
		ifs.close();
		ens_error = (float)num_err/(float)npattern;
		if (save_pred_at_each_iteration) {
		   oresults.seekp(0,ios::beg);
			 oresults << predicted << "-" << i << ends;
			 save_predicted(results, vpredicted, npattern);
		}
		ensemble_errors.push_back(ens_error);
	}
	
	// saving file of the classes predicted by the ensemble at the last iteration
	save_predicted(predicted, vpredicted, npattern);
	
	// saving the ensemble errors at each iteration
	save_ensemble_errors(predicted);
}



void Bagging::save_predicted(char* predicted, const unsigned * vpredicted, unsigned const npattern) {
  
	ofstream fpredicted(predicted);	
	if (fpredicted == NULL)
	  error_exit("Bagging::save_predicted : Can't open file ", predicted);
  for (int j=0; j<npattern; j++)
	  fpredicted << vpredicted[j] << endl;
	if (fpredicted.bad())
	    error_exit("Bagging::save_predicted : Error writing file ", predicted);
	fpredicted.close();
}



/*
	 Computes the error and the confusion matrix of the ensemble.
*/
float Bagging::calc_error_and_confusion(char* testfile, char* predicted, bool save_err) {	
	 	 
	unsigned const SIZE = 256;
	char testlabels[SIZE];
	char model[SIZE];
	int predicted_class;
	int trueclass;
	unsigned npattern;
	float err;
	unsigned num_err = 0;
	
	// Computing the error rate of the ensemble and the confusion matrix
	ostrstream otestlabels(testlabels,SIZE,ios::out);
	otestlabels << testfile << ".labels" << ends;
	ifstream ftest(testlabels);	
	if (ftest == NULL)
	  error_exit("Bagging::calc_error_and_confusion : Can't open file ", testlabels);
	// opening file of the classes predicted by the ensemble
	ifstream fpredicted(predicted);	
	if (fpredicted == NULL)
	  error_exit("Bagging::calc_error_and_confusion : Can't open file ", predicted);
	// opening files of the predicted classes computed by the base learners
	fpredicted.seekg(0,ios::beg);	 
	ftest.seekg(0,ios::beg);
	confusion.initialize();	
	npattern = wc(testlabels);
	for (unsigned i = 0 ; i<npattern; i++) {
	  ftest >> trueclass;
		fpredicted >> predicted_class;
		confusion(predicted_class-1, trueclass-1) += 1;
		if (predicted_class != trueclass)
		  num_err++;
	}
	
	ostrstream omodel(model,SIZE,ios::out);
	err = (float)num_err/(float)npattern;
	if (save_err) {
	  omodel.seekp(0,ios::beg);
		omodel << predicted << SUFFIX_BASE_LEARNER_ERRORS << ends;
		save_base_learner_errors(model, base_learner_test_errors);
	}		
		
	return err;
}	


/*
	 Combines the bagged ensemble of trained base learners using model files.
*/
float Bagging::combine_models (char* testfile, char* predicted, unsigned startiteration,
                              unsigned maxiterations, bool save_err,  bool b_printout) {
  
	float err;	
	if (b_printout)
	  cout << "Testing the base learners ... " << endl;		
	// Tests the trained base learners stored in files
	test_base_learners(testfile, startiteration, maxiterations);
	if (b_printout)
	  cout << "Computing the classes predicted by the ensemble ... " << endl;	
	// Computes the files of the classes predicted by the ensemble
	compute_ensemble_results(testfile, predicted,  startiteration,  maxiterations);
	// Computes the error and the confusion matrix of the ensemble
	if (b_printout)
	  cout << "Computing the error and the confusion matrix ... " << endl;	
	err = calc_error_and_confusion(testfile, predicted, save_err);

  return err;
}	
	
void Bagging::save_base_learner_errors(char* filename, deque<float>& b) {
  ofstream of(filename);
	if (!of)
	  error_exit("Bagging::save_base_learner_errors : Can't open file ", filename);
  for (unsigned i = 0; i < b.size(); i++)
	  of << b[i] << endl;
	if (of.bad())
	  error_exit("Bagging::save_base_learner_errors : Error writing file ", filename);
}

/*
	Saves into a file filename.ens_errors the error rate of the ensemble at each iteration
*/ 
void Bagging::save_ensemble_errors(char* filename) {
  
	unsigned const SIZE = 256;
	char buf[SIZE];
	ostrstream oenserr(buf,SIZE,ios::out);
	oenserr << filename << SUFFIX_ENSEMBLE_ERRORS << ends;
  
	ofstream of(buf);
	if (!of)
	  error_exit("Bagging::save_ensemble_errors : Can't open file ", buf);
  for (unsigned i = 0; i < ensemble_errors.size(); i++)
	  of << ensemble_errors[i] << endl;
	if (of.bad())
	  error_exit("Bagging::save_ensemble_errors : Error writing file ", buf);
}


/*
	 Combines the bagged ensemble of trained base learners using model files storing the results at each iteration.
*/
float Bagging::combine_models_step (char* testfile, char* predicted, unsigned startiteration,
                                    unsigned maxiterations, bool save_err,  bool b_printout, bool b_save_pred_at_each_iteration){
															
  float err;
	
	// Tests the trained base learners stored in files
	if (b_printout)
	  cout << "Testing the base learners on the file " << testfile << " ... " << endl;
	test_base_learners(testfile, startiteration, maxiterations);	
	// Computes the ensemble errors at each iteration and stores in the file predicted 
	//the classes predicted by the ensemble at the last iteration.	
	if (b_printout)
	  cout << "Computing the classes predicted by the ensemble ... " << endl;												
	compute_ensemble_results_step(testfile, predicted, startiteration, maxiterations, b_save_pred_at_each_iteration);
	// Computes the error and the confusion matrix of the ensemble
	if (b_printout)
	  cout << "Computing the error and the confusion matrix ... " << endl;	
	err = calc_error_and_confusion(testfile, predicted, save_err);
  return err;												
}

/*
	Computes by majority voting the class
*/ 
unsigned Bagging::majority_voting(unsigned* vclasses) {
  unsigned predicted;
	unsigned maxx = 0;
	deque<unsigned>maximal_classes;
	
	for (int i=0; i<nclass; i++) {
	  if (vclasses[i] > maxx) {
		  maxx = vclasses[i];
			maximal_classes.clear();
			maximal_classes.push_back(i+1);
		} else if (vclasses[i] == maxx)
		  maximal_classes.push_back(i+1);
	}
	if (maximal_classes.size() == 1)
		  predicted = maximal_classes[0];
	else {
		  long seed = (long) -time(NULL);
			unsigned guess = (unsigned)(ran1(&seed)*maximal_classes.size());
			predicted = maximal_classes[guess];
	}
	return predicted;		
}

/*
	Prints the confusion matrix to an output stream
*/ 
void Bagging::print_conf_matrix(ostream& os){
  
	os << "Confusion matrix of the bagged ensemble:" << endl;
	os.precision(0);
  confusion.print(os);
}

/*
	Prints info about the bagged ensemble
	@param O output stream
*/ 
void Bagging::print_info(ostream& O) {

  unsigned i;
	O << "Bagged ensemble of learning machines" << endl;
	O << "Number of classes = " << nclass << endl;
	O << "Number of base learners = " << baselist.size() << endl;
	if (ensemble_errors.size() > 0) {
		 O << "Ensemble errors at each iteration: " << endl;
		 for (i=0; i < ensemble_errors.size(); i++) 
			 O << i+1 << ": " << ensemble_errors[i] << endl;
	}
	print_conf_matrix(O);
}
	 

