//  "bagsvm.cc"
/*
  Application for training a bagged ensemble of SVM. The base learners are homogeneos SVM. They can be linear, gaussian, polynomial
	or sigmoid. 
	In output is printed the error rate on the training and on the testing set and the confusion matrix.
	In file pred.train and pred.test are stored respectively the predicted classes for the traiing and for the testing set.
	The error of the ensemble at each iteration can be stored ib the files pred.train.ens_errors (train set) and
	pred.test.ens_errors (test set). The error of the base learners can be stored in the files model.blerr (error on bootstrapped data
	sets), pred.train.blerr (whole train set), pred.test.blerr (test set).
*/

#include<stdlib.h>
#include"utils.h"
#include"basesvml.h"
#include"bagging.h"
#include"matrix.h"

static unsigned const SIZE = 256;
static unsigned const DEF_BASE_LEARNERS = 100;
static float const DEF_C = 1.0;
static unsigned const DEF_DEGREE = 2;
static float const DEF_S = 1.0;
static float const DEF_R = 1.0;
static float const DEF_SIGMA = 1.0;
static const char* const BASE_MODEL_NAME = "model";
static const char* const PREDICTED_TRAIN = "predtrain";
static const char* const PREDICTED_TEST = "predtest";
static unsigned type = 0; // SVM type
static unsigned n_base_learners = DEF_BASE_LEARNERS;
static float C = DEF_C;
static unsigned degree = DEF_DEGREE;
static float s = DEF_S;
static float r = DEF_R;
static float sigma = DEF_SIGMA;
static bool b_printout = true;
static bool b_save_base_train_err = false;
static bool b_save_base_test_err = false; 
static bool b_ens_train_err = false;
static bool b_ens_test_err = false;
static bool b_test_on_train = false;
static bool b_save_pred_at_each_iteration = false;

static char trainfile[SIZE];  
static char testfile[SIZE];   
static char basemodel[SIZE]; 
static char predicted_train[SIZE]; 
static char predicted_test[SIZE]; 



///////////////////////////////////  
static char outfile[SIZE];   
static char saverr[SIZE];   

static unsigned savenet = 0;   
static unsigned saveout = 0;   
static int seed = 0;  

 
static void print_info_use(void);

int doarg(char* arg[], int i);

////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  int num_train, i;
	clock_t start, end;
	float err_train, err_test;
	BaseSVM_light_common* base = NULL;
  	
  if (argc < 2) 
     print_info_use();
  if ((argc%2) == 1)
     print_info_use();
		 
	strcpy(trainfile, argv[1]);
	strcpy(testfile, argv[1]);
	strcpy(basemodel, BASE_MODEL_NAME);
	strcpy(predicted_train, PREDICTED_TRAIN);
	strcpy(predicted_test, PREDICTED_TEST);
	
	for (i=2; i<argc; i+=2) 
     if (doarg(argv, i))
        print_info_use();
	
  // print general info to stdout
	cout << "Bagged SVM ensemble" << endl;
	cout << "Number of base learners: " << n_base_learners << endl;
  cout << "Number of classes: 2" << endl;
	cout << "Base SVM type : ";
	switch (type) {
	  case 0: cout << "Linear" << endl;
		        cout << "Parameter C = " << C << endl; 
		        base = new BaseSVM_light_linear (C);
		        break;
		case 1: cout << "Polynomial" << endl;
		        cout << "Parameters:" << endl;
						cout << "  degree = " << degree << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl; 
						base = new BaseSVM_light_poly (degree, C, s, r);
						break;
		case 2: cout << "RBF (gaussian)" << endl;
		        cout << "Parameters:" << endl;
						cout << "  sigma = " << sigma << endl;
						cout << "  C = " << C << endl; 
						base = new BaseSVM_light_gaussian (1.0/sigma, C);
						break;
		case 3: cout << "Sigmoid" << endl;
		        cout << "Parameters:" << endl;
						cout << "  s = " << s << endl;
						cout << "  r = " << r << endl;						
						cout << "  C = " << C << endl;
						base = new BaseSVM_light_sigmoid (C, s, r);
						break;
    default:  cout << "Not known SVM kernel type" << endl;
	            exit(1); 
	}
  	
	if (base == NULL)
	  error_exit(argv[0],  " : Error creating base SVM");
	
	// building the bagged ensemble of SVM
	Bagging bagSVM(base);

	start = clock();
	// training of the bagged SVM ensemble
	bagSVM.train_base_model (trainfile, basemodel, 1, n_base_learners, b_save_base_train_err, b_printout);
	
	
  if (!strcmp(trainfile,testfile)) { // test only on the train set if it is equal to the test set  
	  if (b_save_pred_at_each_iteration) 
	    err_test = err_train = bagSVM.combine_models_step (trainfile, predicted_train, 1, n_base_learners, 
		                                                     false, b_printout, b_save_pred_at_each_iteration);
		else {
      if (!b_ens_train_err)
		    err_test = err_train = bagSVM.combine_models (trainfile, predicted_train, 1, n_base_learners, false, b_printout);
		  else
		    err_test = err_train = bagSVM.combine_models_step (trainfile, predicted_train, 1, n_base_learners, false, b_printout);
		}
	}
	else {
	  if (b_test_on_train)  // test on the train set
		  if (b_save_pred_at_each_iteration) 
	    err_train = bagSVM.combine_models_step (trainfile, predicted_train, 1, n_base_learners, 
		                                          false, b_printout, b_save_pred_at_each_iteration);
		  else {
		  	 if (!b_ens_train_err)
		    	 err_train = bagSVM.combine_models(trainfile, predicted_train, 1, n_base_learners, b_save_base_train_err, b_printout);
		  	 else
		    	 err_train = bagSVM.combine_models_step(trainfile, predicted_train, 1, n_base_learners, b_save_base_train_err, b_printout);
		  }
		bagSVM.ensemble_errors.clear();
		if (b_save_pred_at_each_iteration) 
	    err_test = bagSVM.combine_models_step (testfile, predicted_test, 1, n_base_learners, 
		                                          false, b_printout, b_save_pred_at_each_iteration);
		else {		
		  if (!b_ens_test_err)  // test on the test set
		     err_test = bagSVM.combine_models(testfile, predicted_test, 1, n_base_learners, b_save_base_test_err, b_printout);
		  else
		     err_test = bagSVM.combine_models_step(testfile, predicted_test, 1, n_base_learners, b_save_base_test_err, b_printout);
		}
	}  
	end = clock();
	
	bagSVM.print_conf_matrix(cout);
	if (b_test_on_train) 
	   cout << endl << "Ensemble error on the training set = " << err_train << endl;
	cout <<  "Ensemble error on the testing set = " << err_test << endl;
	print_elapsed_cpu_time(start, end);
	
  delete base;
  return 0;
}


void print_info_use(void)
{
  cerr << "
  bagsvm. NEURObjects 2.0 application for training and testing
          a bagged ensemble of SVM.
  Usage: bagsvm trainfile [options]
  Options:
  -test string        test file name
  -type unsigned      SVM base learner type:
                      0: linear (default)
                      1: polynomial
                      2: RBF (gaussian)
                      3: sigmoid (MLP-like)
  -nb unsigned        number of base learners  
  -c float            regularization parameter of the base SVM (def. 1)
  -deg unsigned       degree of the polynomial SVM (s a*b + r)^d (def. 2)
  -s float            s parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -r float            s parameter of the poly SVM or of the sigmoid SVM (def. 1)
  -sigma float        sigma parameter of the gaussian SVM: exp(-(a-b)^2/sigma)
  -sbmodel string     save the learned base models onto sbmodel1..sbmodeln	
  -sbtrain unsigned   save training errors of the base learners: 0 (def) no, 1 yes
  -sbtest  unsigned   save testing errors of the base learners: 0 (def) no, 1 yes
  -setrain unsigned   save ensemble train errors at each step: 0 (def) no, 1 yes
  -setest  unsigned   save ensemble test errors at each step: 0 (def) no, 1 yes
  -pred  string       save classes predicted by the ensemble 
                      with respect to the test set onto a file.
  -predtrain  string  save classes predicted by the ensemble 
                      with respect to the train set onto a file.
  -sit unsigned       if 1 save save classes predicted by the ensemble
                      onto a file at each iteration (def. 0)
  -train unsigned     if 0 (def.) testing is not performed on train set, if 1 yes
  -p integer          1 = printing a point at each iteration; 0 = noprint (def.)
  " << endl;
  exit(1);
}


int doarg(char* arg[], int i)
{
  int j;
	if (!strcmp(arg[i], "-test"))
    strcpy(testfile, arg[i+1]);
	else if (!strcmp(arg[i], "-type"))
    type = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-nb")) 
    n_base_learners = atoi(arg[i+1]);
  else if (!strcmp(arg[i], "-c")) 
    C = atof(arg[i+1]);
  else if (!strcmp(arg[i], "-deg")) 
    degree = atoi(arg[i+1]);
	else if (!strcmp(arg[i], "-s")) 
    s = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-r")) 
    r = atof(arg[i+1]);
	else if (!strcmp(arg[i], "-sigma")) {
    sigma = atof(arg[i+1]);
		if (sigma == 0) sigma = 1.0; }
	else if (!strcmp(arg[i], "-sbmodel")) 
    strcpy(basemodel, arg[i+1]);
	else if (!strcmp(arg[i], "-sbtrain")) {
    if (atoi(arg[i+1]) == 1) b_save_base_train_err = true; }
	else if (!strcmp(arg[i], "-sbtest")) {
    if (atoi(arg[i+1]) == 1) b_save_base_test_err = true; }
	else if (!strcmp(arg[i], "-setrain")) {
    if (atoi(arg[i+1]) == 1) b_ens_train_err = true; }
	else if (!strcmp(arg[i], "-setest")) {
    if (atoi(arg[i+1]) == 1) b_ens_test_err = true;} 
  else if (!strcmp(arg[i], "-pred"))  
		strcpy(predicted_test, arg[i+1]);
  else if (!strcmp(arg[i], "-predtrain"))  
    strcpy(predicted_train, arg[i+1]);
	else if (!strcmp(arg[i], "-train")) {
    if (atoi(arg[i+1]) == 1) b_test_on_train = true; }
	else if (!strcmp(arg[i], "-p")) {
    j = atoi(arg[i+1]);
		if (j == 0)
		   b_printout = false;}
	else if (!strcmp(arg[i], "-sit")) {
    j = atoi(arg[i+1]);
		if (j == 1)
		   b_save_pred_at_each_iteration = true;}
  else 
    return 1;
  return 0;
} 
