/* ******************************************************************************
NEURObjects 2.0
baselearner.h January 2002
Giorgio Valentini
****************************************************************************** */
#ifndef BASELEARNER_H
#define BASELEARNER_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include "tset.h"

using namespace std;

static char* const result_file_base = "o";

/**
   Virtual abstract class of the base learners with basic functionalities.\\
	 This class is a virtual abstract class of the base learners used in ensemble classes as Bagging
	 and WeightedBagging.\\
	 This class defines the interface of the base learners used in ensemble algorithms. 
	 Each real istanceable base learner must inherit from this base class and implement the methods of
	 this interface.\\
	 No assumptions are made about the way the learning is performed by the base learner. In such a
	 way you can wrap inside the derived base learners existing learner classes or also learning
	 algorithms previously implemented, in order to use them in ensemble algorithms.
	 Each implemented base learner must have a training method that uses a file of data to learn 
	 (no assumptions are made about the format of the data) and a model file to store the learned
	 model (no assumptions are made about the format of the model file). In addition the base learner
	 store the computed results in a file. This file is used by the ensemble class to combine the
	 results of the base learners.
	 Each base learner is identified by a number that must be unique in a given ensemble. This number corresponds
	 to the order in the deque list storing the base learners.
*/ 
class BaseLearner {

protected:

unsigned num;

public:
// members

// methods

/// set the number of the base learner
void set_number (unsigned n) {num = n;};

/// get the number of the base learner
unsigned get_number (void) const {return num;};

/**
  Method that produces a clone of the current learning machine.\\
	@return pointer to the cloned learning machine
*/
virtual BaseLearner* clone() = 0;

/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. No predeterminated format is required both for the training set and model files.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model) = 0;
											


/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a file.
	 The output results (computed classes) are stored in a file.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by the ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
virtual float test_learner(char* testfile, char* model, char* predicted) = 0;


};


//////////////////////////////////////////////////////////////////////////////////////////////////
/**
   Virtual abstract class of the base learners with extended functionalities.\\
	 This class is a virtual abstract class of the base learners used in ensemble classes.\\
	 This class defines the interface of the base learners used in ensemble algorithms. 
	 Each real istanceable base learner must inherit from this base class and implement the methods of
	 this interface.\\
	 No assumptions are made about the way  learning is performed by the base learner, but the 
	 BaseLearnerExt uses the TrainingSet class to hold the data and must be capable to maintain the learned model
	 in memory. Tipically NEURObjects learners can be used as BaseLearnerExt. Use instead derived BaseLearner
	 classes to wrap learning algorithms that uses particular data format and that need to store in files the
	 learned model.
	 Each base learner is identified by a number that must be unique in a given ensemble. This number corresponds
	 to the order in the deque list storing the base learners.
*/ 
class BaseLearnerExt : public BaseLearner {


public:

/// vector of the predicted classes
deque<unsigned> predicted_classes;

/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. No predeterminated format is required both for the training set and model files.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model) = 0;

/**
	 Trains the base learner.\\
	 The base learner is trained using a training set stored in memory through an instance of the TrainingSet
	 class. The learned model can be  stored in the file model. In both cases the learned model is maintained in
	 memory ready for use by the method test.
	 @param Tset  name of the training set
	 @param model  model file name. If set to NULL no model file is stored	 
	 @return error rate
*/
virtual float train_learner (TrainingSet& Tset, char* model = NULL) = 0;
											
/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a file.
	 The output results (computed classes) are stored in a file.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by hte ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
virtual float test_learner(char* testfile, char* model, char* predicted) = 0;

/**
	 Tests the base learner.\\
	 The learner is tested using a testing set stored in memory through an instance of the TrainingSet
	 class.
	 The output results (computed classes) must be stored in the deque container predicted_classes. 
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by the ensemble classes to
	 combine the outputs of the base learners.	 
	 @param Tset  name of the testing set.
	 @param model  model file name. If different from NULL a model file is loaded.
	 @param  predicted computed class file. If set to NULL no predicted file is stored	 
	 @return error rate
*/
virtual float test_learner(TrainingSet& Tset, char* model= NULL, char* predicted = NULL) = 0;


/**
	 Computes the predicted class on a specific input.\\
	 @param v input vector.
	 @return predicted class
*/
virtual unsigned test_pattern (vect& v) = 0;

/**
	 Save the learned model into a file.\\
	 @param model  model file name
*/
virtual void save_model (char* model) = 0;

};



#endif
