/* ******************************************************************************
NEURObjects 2.0
basemlp.h January 2002
Giorgio Valentini
****************************************************************************** */
#ifndef BASEMLP_H
#define BASEMLP_H

#include <iostream>
#include <fstream>
#include <string>
#include "baselearner.h"
#include "net.h"

using namespace std;


//////////////////////////////////////////////////////////////////////////////////////////////////


/**
   Abstract base class for MLP base learners.\\
	 This class collects the members and methods common to the base learners implementing Multi Layer
	 Perceptrons. The derived classes are used in bagging and boosting ensemble classes.
	 Learning is performed reading a training file in NEURObjects format or reading an instance of a
	 TrainingSet class. The corresponding learned model can be maintained in memory or stored in a
	 file. The classification results can be stored  in a file according to the BaseLearner
	 class format or mainntained in memory. Moreover a file with suffix .out containing the computed coninuous 
	 outputs of the MLP is generated.
	 
*/ 
class BaseMLP_common : public BaseLearnerExt {

protected:

/// suffix of the files containing continuous outputs computed by the MLP
static char* const OUT = ".out";

/// Default separator used in the data sets
static char const DEFAULT_SEPARATOR = ',';

/// Separator used in the data sets
char separator;

/// Maximun number of itrations allowed.
unsigned max_iterations;

/// Normalized RMS error theshold
double RMS_threshold;
 
/// Learning algorithm
Learning* learn;

/// Vector of the mean of the data file
vect mean;

/// Vector of the standard deviation of the data file
vect stdev;

public:

/// Get the max_iterations parameter
unsigned const get_max_iterations() const {return max_iterations;}

/// Set the max_iterations parameter
void set_max_iterations(unsigned val) {max_iterations=val;}

/// Get the RMS_threshold parameter
double const get_RMS_threshold() const {return RMS_threshold;}

/// Set the RMS_threshold parameter
void set_RMS_threshold(double val) {RMS_threshold=val;}

/// Get the Learning algorithm
Learning* const get_learning_algorithm() const {return learn;}

/// Set the Learning algorithm
void set_learning_algorithm(Learning* l) {learn=l;}

/// Get the mean
vect& get_mean()  {return mean;}

/// Set the mean
void set_mean(vect v) {mean=v;}

/// Get the stdev
vect&  get_stdev() {return stdev;}

/// Set the stdev
void set_stdev(vect v) {stdev=v;}

/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a
	 file, both in SVM-light format.
	 It outputs two files: the predicted classes file in the format specified in the base class BaseLearner,
	 and a file with the same and an added suffix .out that stores the continuous outputs of the SVM.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by hte ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
virtual float test_learner(char* testfile, char* model, char* predicted) = 0;

};



//////////////////////////////////////////////////////////////////////////////////////////////////

/**
   One layer perceptron base learner.\\
	 This class implements a MLP bas learner with o hidden layer, inheriting both from BaseMLP_common and
	 OneLayer.
	 
*/ 
class BaseMLP_OneLayer : public BaseMLP_common, public OneLayer {



public:

/// Dummy constructor
BaseMLP_OneLayer();

/** Constructor. \\
    In the constructor you must specify the number of inputs, outputs, the learning algorithm, and
		the stopconditions. that is the mormalized RMS error and the maximum nimber of iterations
		allowed for the backpropagation algorithm. Note that the learning algorithm to be used by the
		base learner must be built before the calling to the constructor.
		@param nout number of outputs
		@param nin number of inputs
		@learn_alg learning algorithm
		@param errtol normalized RMS error threshold
		@param maxits maximun number of iterations allowed
		@param n order number of the base learner   
		@param sep separator used in the data sets
		@param s intialization seed for the weights. If s=0 a random seed is generated
*/
BaseMLP_OneLayer (unsigned nout, unsigned nin, Learning* learn_alg, double errtol = ERR_TOL, 
                  unsigned maxits = MAXITS, unsigned n = 0, char sep = DEFAULT_SEPARATOR, int s = 0);

/// Destructor
virtual ~BaseMLP_OneLayer () {}



/**
	 Trains the base learner using a training set stored in a file.\\
	 The base learner is trained using the data set trainfile. The learned model is stored in the
	 file model. The file formats are in  NEURObjects file format.
	 @param trainfile  name of the training set
	 @param model  model file name.
	 @return error rate
*/
virtual float train_learner (char* trainfile, char* model);

/**
	 Trains the base learner.\\
	 The base learner is trained using a training set stored in memory through an instance of the TrainingSet
	 class. The learned model can be  stored in the file model. In both cases the learned model is maintained in
	 memory ready for use by the method test.
	 @param Tset  name of the training set
	 @param model  model file name. If set to NULL no model file is stored	 
	 @return error rate
*/
virtual float train_learner (TrainingSet& Tset, char* model = NULL){}
											

/**
	 Tests the base learner using a testing set stored in a file.\\
	 The learner is tested using a testing set stored in a file and a  previously learned model  stored in a file.
	 The output results (computed classes) are stored in a file.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by hte ensemble classes to
	 combine the outputs of the base learners.	 
	 @param testfile  name of the testing set. 
	 @param model  model file name.
	 @param predicted computed class file.
	 @return error rate
*/
virtual float test_learner(char* testfile, char* model, char* predicted);




/**
	 Tests the base learner.\\
	 The learner is tested using a testing set stored in memory through an instance of the TrainingSet
	 class.
	 The output results (computed classes) must be stored in the vector predicted_classes. The method must
	 allocate on the heap with new the vector of predicted classes. The destructor must deallocate it.
	 Output file format: The output classes are stored one for each line. The classes must be labeled by
	 integers: The first class by 1, the second by 2, etc. The output files are used by the ensemble classes to
	 combine the outputs of the base learners.	 
	 @param Tset  name of the testing set.
	 @param model  model file name. If different from NULL a model file is loaded.
	 @param  predicted computed class file. If set to NULL no predicted file is stored	 
	 @return error rate
*/
virtual float test_learner(TrainingSet& Tset, char* model= NULL, char* predicted = NULL){}


/**
	 Computes the predicted class on a specific input.\\
	 @param v input vector.
	 @return predicted class
*/
virtual unsigned test_pattern (vect& v){}
											
};


















#endif
