/*
--------------------------------------------------------------------------------

  decomp.h
    
--------------------------------------------------------------------------------
*/  


#ifndef SCOMP_H
#define SCOMP_H

#include<iostream.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include<vector.h>
#include"act_fun.h"
#include"utils.h"
#include"matrix.h"

/**
  One-Per-Class (OPC) decomposition.\\
	@memo OPC decomposition class - HTML documentation under construction
*/
class d_OPC {

protected:
  
	/// Number of classes, i.e. number of codewords
  unsigned n_class;
  
  /// Number of dichotomic tasks, i.e. codeword length
  unsigned n_dico;  
  
  /// Decomposition matrix
  matrix D;
  
public:
  
  enum matrix_val {LOW = -1, IGNORE = 0, HIGH = 1};
  
  
  d_OPC (){}
  
  d_OPC (unsigned nclass, unsigned ndico=0);
	
	d_OPC (char * dec);
  
  d_OPC(const d_OPC& d);
  
  d_OPC& operator=(const d_OPC& d);
  
  ~d_OPC() {}
  
  
  matrix& Read_matrix(void){return D;}
	
	
  matrix Copy_matrix(void){return D;}
  
  
  void save_matrix(ofstream& f);
	
	unsigned Hamming_distance(const vect& v1, const vect& v2);
	
  unsigned MinHamming(unsigned nclass = 3);
	
	float Avg_Hamming_distance(void);
	
	unsigned Min_Dichotomy_distance(void);
	
	float Avg_Dichotomy_distance(void);
	
	bool Banal_dichotomy(void);
    
  unsigned Read_n_class(void) {return n_class;}
   
  unsigned Read_n_dico(void) {return n_dico;}
  
  
 
  void print_matrix(void);
  
  
  //save_matrix (ofstream& ofs);
  
 
 
  
protected:
  
  void Init_D(matrix_val val);
  
};



/**
  Pair-Wise_Coupling (PWC) decomposition.\\ 
  @memo PWC decomposition class - HTML documentation under construction
*/
class d_PWC : public d_OPC {

private:
  
  void Do_PWC_matrix(void);
  
public:
  
  d_PWC(){}
  
  d_PWC (unsigned nclass, unsigned ndico = 0);
  
  d_PWC (char* file);
	
  d_PWC(const d_PWC& d);
  
  d_PWC& operator=(const d_PWC& d);
  
  
};





/**
  Correcting Classifiers Pair-Wise_Coupling (CC) decomposition.\\ 
  @memo CC decomposition class - HTML documentation under construction
*/
class d_CC : public d_OPC {

private:
  
  void Do_CC_matrix(void);
  
public:
  
  d_CC(){}
  
  d_CC (unsigned nclass, unsigned ndico = 0);
  
	d_CC (char* file);
  
  d_CC(const d_CC& d);
 
  d_CC& operator=(const d_CC& d);
  
};




/**
  Error Correcting Output Coding (ECOC) decomposition, exhaustive algorithm.\\
	@memo ECOC decomposition class (exahustive) - HTML documentation under construction
*/
class d_ECOC_ex : public d_OPC {

private:
  
  void Do_ECOC_matrix_exaustive(void);
   
  void Append(unsigned col, unsigned rig, unsigned n, matrix_val val);
  
public:
  
  d_ECOC_ex(){}
  
  d_ECOC_ex (unsigned nclass, unsigned ndico = 0);
  
  d_ECOC_ex (char* file);
	
  d_ECOC_ex(const d_ECOC_ex& d);
  
  d_ECOC_ex& operator=(const d_ECOC_ex& d);
  

};






/**
  Error Correcting Output Coding (ECOC) decomposition, BCH algorithm.\\ 
	@memo ECOC decomposition class (BCH) - HTML documentation under construction
*/
class d_ECOC_BCH : public d_OPC {

private:
  /// data dimension for index_of, alpha_to
  static	const int MAX_DIM_DATA = 10485;
  /// dimension for polinomial q and bb
  static	const int MAX_DIM_POLI = 5485;
  /// order of the Galois field GF(2**m) 
  int  m;
	/// n = 2**m - 1 = size of the multiplicative group of GF(2**m)
	int n;
	///  k = n - deg(g(x)) = dimension (no. of information bits/codeword) of the code
	int k;
	/// t = error correcting capability (max. no. of errors the code corrects)
	int t;
	/// d = 2*t + 1 = designed min. distance = no. of consecutive roots of g(x) + 1
	int	d;  
	/// p[] = coefficients of a primitive polynomial used to generate GF(2**m)
  int  p[21];
	/// alpha_to [] = log table of GF(2**m) 
  int alpha_to[MAX_DIM_DATA];
	/// index_of[] = antilog table of GF(2**m)
	int  index_of[MAX_DIM_DATA];
	///  g[] = coefficients of the generator polynomial, g(x)
	int  g[MAX_DIM_POLI];
	///  recd[] = coefficients of the received polynomial 
  int  recd[MAX_DIM_DATA];
	///  data[] = information bits = coefficients of data polynomial, i(x)
	int  data[MAX_DIM_DATA];
	///  bb[] = coefficients of redundancy polynomial x^(length-k) i(x) modulo g(x)
	int  bb[MAX_DIM_POLI];

  
  void Do_ECOC_matrix_BCH(void);
	
  /**
 		Compute m, order of the Galois field GF(2**m).
		m is the degree of a primitive polynomial p(x) used to compute the
 		Galois field GF(2**m) on the basis of the length of the codeword. 
  	Get precomputed coefficients p[] of p(x).
		@param length length of the binary codeword (number of dichotomizers) 
  */
  void compute_p(const int length);
	
	/**
  	Generate field GF(2**m).
		Generate field GF(2**m) from the irreducible polynomial p(X) with
  	coefficients in p[0]..p[m].\\ 
  	Lookup tables:\\
  	index->polynomial form: alpha_to[] contains j=alpha^i; \\
  	polynomial form -> index form:	index_of[j=alpha^i] = i \\ 
  	alpha=2 is the primitive element of GF(2**m) 
  */
  void generate_gf();
	
	/**
		Compute the generator polynomial of a binary BCH code. 
		Fist generate the
		cycle sets modulo 2**m - 1, cycle[][] =  (i, 2*i, 4*i, ..., 2^l*i). Then
		determine those cycle sets that contain integers in the set of (d-1)
		consecutive integers {1..(d-1)}. The generator polynomial is calculated
		as the product of linear factors of the form (x+alpha^i), for every i in
		the above cycle sets.
  */
  void gen_poly(const int n_class);
		
	/**
		Compute redundacy bb[], the coefficients of b(x). 
		The redundancy polynomial b(x) is the remainder after dividing x^(n_dico-k)*data(x)
		by the generator polynomial g(x).
	*/
	void encode_bch();
	
	/**
  	Translate a decimal number into a binary one.
		@paran num decimal number
		@param bin bunary number expressed as vector of integer
		@param len number of digits of the binary number
  */
  void DecToBin(int num, int bin[], int len);

	
  int equalrow(int r1, int r2);

public:
  
  d_ECOC_BCH(){}
  
  d_ECOC_BCH (unsigned nclass, unsigned ndico);
  
  d_ECOC_BCH (char* file);
	
  d_ECOC_BCH(const d_ECOC_BCH& d);
  
  d_ECOC_BCH& operator=(const d_ECOC_BCH& d);
  
};




/**
  Decomposition random algorithm with greater or equal constraints.\\ 
	It generates random constrained codes using random generation and selecting the  codes with the maximum of the minimum pairwise
	Hamming distance between the codewords above a desired threshold, or, if the maximun number of iterations is reached, 
	the best codes generated.
	Only correct and not trivial decomposition matrices are generated (that is only decompostion matrices with no equal rows or columns
	without complementary rows). 
	Generation algorithms allow constraints also on the minimum Hamming distance between rows (diversity of dichotomizers'
	task).
  @memo Random decompsition algorithm with greater or equal constraints.
*/
class d_Rand_gr_equal_constraints : public d_OPC {

protected:
  /// Iterations of the random generation
  unsigned times;
	
	/// initial seed for the random generator
	long seed;
	
  /**
   Build a decomposition matrix D using random codes and returns the minimum Hamming distance between columns.
	 It tries ntimes times to generate randomly the codewords, considering valid a guess only if the minimum Hamming 
	 distance between rows (dichotomies) is above min_dist, retaining the bst matrix (i.e. the matrix with maximum
	 pairwise minimum distance among columns (codewords).
	 @param min_dicho_dist minimum Hamming distance between each pair od rows
	 @param  min_codeword_dist maximum of the minimum Hamming distance between each pair of columns (codewords)
	 @param ntimes number of random guessing of the decomposition matrix
	 @return minimum hamming distance between columnd of the selected deomposition matrix
  */  
  virtual unsigned Do_Rand_matrix(unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 2, unsigned ntimes = 1);
  
	/**
   Build a single decomposition matrix D using random codes.
	 It generates a random matrix with minimum Hamming distance between rows equal to min_dist. It generates only non trivial
	 decompositiom matrices.
	 @param min_dicho_dist minimum Haming distance between each pair od rows
  */  
  void Make_Rand_matrix(unsigned min_dicho_dist = 1);
	
	/**
   Generates a random row of the matrix with a fixed number of +1 and -1.\\
	 @param x vector to be generated
	 @param  pos if false a row of -1 is generated except for random n_pos positions, otherwise the opposite.
	 @param n_pos number of random +1 (if pos is false) or -1 (if pos is true)
	 @return vector to be generated
  */  
	vect& generate_row(vect& x, bool pos = false, int n_pos = 1);
	
	
public: 

  // Dummy constructor
  d_Rand_gr_equal_constraints(){}
  /** Constructor.\\
     @param nclass number of classes (columns of the random decomposition matrix)
	   @param ndico  number of dichotomies (rows of the random decomposition matrix)
		 @param min_dicho_dist minimum Hamming distance between each pair of rows
		 @param  min_codeword_dist minimum Hamming distance between each pair of columns (codewords)
	   @param ntimes number of random guessing of the decomposition matrix
  */
  d_Rand_gr_equal_constraints (unsigned nclass, unsigned ndico, unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 2,
	                             unsigned ntimes = 10);
	
	/** Constructor.\\
     @param file file where the decomposition matrix is stored
  */
	d_Rand_gr_equal_constraints (char* file);
	
	/// Destructor
	virtual ~d_Rand_gr_equal_constraints() {}
	
	/**
   Algorithm to decrement the minimum Hammin distance between codewords.\\
	 It tries to decremet the minimum Hamming distance between codewords substituting randomly a row
	 with a fixed numer of +1 or -1 to the decomposition matrix. Note that if the attempt is not successfull the 
	 method returns a NULL pointer. 
	 @param n_pos number of random +1 (if pos is false) or -1 (if pos is true)
	 @param ntimes number of random guessing: it is equal to ntimes*rows*columns
	 @param  pos if false a row of -1 is generated except for random n_pos positions, otherwise the opposite.
	 @return the random decomposition matrix if the attemp is successfull, NULL otherwise.
  */  
	d_Rand_gr_equal_constraints * decrementHD(int n_pos = 1, int ntimes = 1, bool pos = false);
  
  /**
   Algorithm to increment the minimum Hammin distance between codewords.\\
	 It tries to increment the minimum Hamming distance between codewords substituting randomly a row
	 with a fixed numer of +1 or -1 to the decomposition matrix. Note that if the attempt is not successfull the 
	 method returns a NULL pointer. 
	 @param n_pos number of random +1 (if pos is false) or -1 (if pos is true)
	 @param ntimes number of random guessing: it is equal to ntimes*rows*columns
	 @param  pos if false a row of -1 is generated except for random n_pos positions, otherwise the opposite.
	 @return the random decomposition matrix if the attemp is successfull, NULL otherwise.
  */  
	d_Rand_gr_equal_constraints * incrementHD(int n_pos = 1, int ntimes = 1, bool pos = false);
  
  ///Copy constructor
  d_Rand_gr_equal_constraints(const d_Rand_gr_equal_constraints& d);
  ///Overload assignement operator
  d_Rand_gr_equal_constraints& operator=(const d_Rand_gr_equal_constraints& d);
  
};


// ---------------------------------------------------------------------------------------------------------------------

/**
  Decomposition random algorithm with  equal constraints.\\ 
	It generates random constrained codes using random generation and selecting the  codes with the maximum of the minimum pairwise
	Hamming distance between the codewords equal to a desired threshold, or, if the maximun number of iterations is reached, 
	the minimum of the maximum.
	Only correct and not trivial decomposition matrices are generated (that is only decompostion matrices with no equal rows or columns
	without complementary rows). 
	Generation algorithms allow constraints also on the minimum Hamming distance between rows (diversity of dichotomizers'
	task).
  @memo Random decomposition algorithm with  equal constraints.
*/
class d_Rand_equal_constraints : public d_Rand_gr_equal_constraints {

protected:
  /**
   Build a decomposition matrix D using random codes and returns the minimum Hamming distance between columns.
	 It tries ntimes times to generate randomly the codewords, considering valid a guess only if the minimum Hamming 
	 distance between rows (dichotomies) is above min_dist, retaining the bst matrix (i.e. the matrix with maximum
	 pairwise minimum distance among columns (codewords).
	 @param min_dicho_dist minimum Hamming distance between each pair od rows
	 @param  min_codeword_dist minimum Hamming distance between each pair of columns (codewords)
	 @param ntimes number of random guessing of the decomposition matrix
	 @return minimum hamming distance between column of the selected deomposition matrix
  */  
  virtual unsigned Do_Rand_matrix(unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 2, unsigned ntimes = 1);
	
public:

  // Dummy constructor
  d_Rand_equal_constraints(){}
  /** Constructor.\\
     @param nclass number of classes (columns of the random decomposition matrix)
	   @param ndico  number of dichotomies (rows of the random decomposition matrix)
		 @param min_dicho_dist minimum Hamming distance between each pair of rows
		 @param  min_codeword_dist maximum of the minimum Hamming distance between each pair of columns (codewords)
	   @param ntimes number of random guessing of the decomposition matrix
  */
  d_Rand_equal_constraints (unsigned nclass, unsigned ndico, unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 2,
	                          unsigned ntimes = 10);
	
	/** Constructor.\\
     @param file file where the decomposition matrix is stored
  */
	d_Rand_equal_constraints (char* file);
  
  ///Copy constructor
  d_Rand_equal_constraints(const d_Rand_equal_constraints& d);
  ///Overload assignement operator
  d_Rand_equal_constraints& operator=(const d_Rand_equal_constraints& d);
  
};


// ---------------------------------------------------------------------------------------------------------------------

/**
  Decomposition random algorithm without constraints.\\ 
	It generates random constrained codes using random generation.
	Only correct and not trivial decomposition matrices are generated (that is only decompostion matrices with no equal rows or columns
	without complementary rows). 
	Generation algorithms allow constraints also on the minimum Hamming distance between rows (diversity of dichotomizers'
	task).
  @memo Random decomposition algorithm without constraints.
*/
class d_Rand : public d_Rand_gr_equal_constraints {

protected:
  /**
   Build a decomposition matrix D using random codes and returns the minimum Hamming distance between columns.
	 @param min_dicho_dist minimum Hamming distance between each pair od rows
	 @param  min_codeword_dist minimum Hamming distance between each pair of columns (codewords)
	 @param ntimes number of random guessing of the decomposition matrix
	 @return minimum hamming distance between column of the selected deomposition matrix
  */  
  unsigned Do_Rand_matrix(unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 1, unsigned ntimes = 1);
	
public:

  // Dummy constructor
  d_Rand(){}
  /** Constructor.\\
     @param nclass number of classes (columns of the random decomposition matrix)
	   @param ndico  number of dichotomies (rows of the random decomposition matrix)
		 @param min_dicho_dist minimum Hamming distance between each pair of rows
  */
  d_Rand (unsigned nclass, unsigned ndico, unsigned min_dicho_dist = 1);
	
	/** Constructor.\\
     @param file file where the decomposition matrix is stored
  */
	d_Rand (char* file);
  
  ///Copy constructor
  d_Rand(const d_Rand& d);
  ///Overload assignement operator
  d_Rand& operator=(const d_Rand& d);
  
};





// ---------------------------------------------------------------------------------------------------------------------

/**
  Decomposition random algorithm generating multiple decomposition matrices.\\ 
	It generates multiple random constrained codes in a selected range.
	Only correct and not trivial decomposition matrices are generated (that is only decompostion matrices with no equal rows or columns
	without complementary rows). 
	Generation algorithms allow constraints also on the minimum Hamming distance between rows (diversity of dichotomizers'
	task).
  @memo Random decomposition algorithm with  equal constraints.
*/
class d_Rand_constraints_multiple : public d_Rand_gr_equal_constraints {

protected:
  /**
   Build a decomposition matrix D using random codes and returns the minimum Hamming distance between columns.
	 It tries ntimes times to generate randomly the codewords, considering valid a guess only if the minimum Hamming 
	 distance between rows (dichotomies) is above min_dist, retaining the bst matrix (i.e. the matrix with maximum
	 pairwise minimum distance among columns (codewords).
	 @param min_dicho_dist minimum Hamming distance between each pair od rows
	 @param  min_codeword_dist minimum Hamming distance between each pair of columns (codewords)
	 @param ntimes number of random guessing of the decomposition matrix
	 @return minimum hamming distance between column of the selected deomposition matrix
  */  
  void d_Rand_constraints_multiple::Do_Rand_matrix_multiple (unsigned min_dicho_dist, unsigned min_codeword_dist, 
                                                         unsigned max_codeword_dist, unsigned ntimes, char *file);
	
public:

  // Dummy constructor
  d_Rand_constraints_multiple(){}
  /** Constructor.\\
     @param nclass number of classes (columns of the random decomposition matrix)
	   @param ndico  number of dichotomies (rows of the random decomposition matrix)
		 @param min_dicho_dist minimum Hamming distance between each pair of rows
		 @param  min_codeword_dist minimum of the minimum Hamming distance between each pair of columns (codewords)
     @param  max_codeword_dist maximum of the minimum Hamming distance between each pair of columns (codewords)
	   @param ntimes number of random guessing of the decomposition matrix
  */
  d_Rand_constraints_multiple (unsigned nclass, unsigned ndico, char *file, unsigned min_dicho_dist = 1, unsigned min_codeword_dist = 2, 
                               unsigned max_codeword_dist = 4, unsigned ntimes = 10);
	
	/** Constructor.\\
     @param file file where the decomposition matrix is stored
  */
	d_Rand_constraints_multiple (char* file);
  
  ///Copy constructor
  d_Rand_constraints_multiple(const d_Rand_constraints_multiple& d);
  ///Overload assignement operator
  d_Rand_constraints_multiple& operator=(const d_Rand_constraints_multiple& d);
  
};










#endif  
  
  
  
  


  
  
  
