/* ******************************************************************************
NEURObjects 2.01
lobag.h November-December 2002
Giorgio Valentini
****************************************************************************** */
#ifndef LOBAG_H
#define LOBAG_H

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include <cstdlib>
#include <cstring>
#include "matrix.h"
#include "tset.h"
#include "baselearner.h"
#include "basesvml.h"
#include "bagging.h"
#include "bootstrapLight.h"
#include "bv2decomp.h"

using namespace std;

// base name of the boostrapped data
static char* const BS_NAME = "bs";

// base name of the models
static char* const MODEL_NAME = "m";

// default directory for the bootstrapped samples
static  char* const DIRBS = "tmpbs";

// default directory for the models
static char* const DIRMODELS = DIRBS;

// default directory for the results computed by the base learners
static  char* const DIR_RES = DIRBS;

struct par_model {
  int type;
  float C;
	int degree;
	float s;
	par_model(float x=1, int t=0) { type=t; C=x; degree = (int)x; s=x;}
	void print (ostream& out = cout);
};

struct model_BV {
  par_model P;
	AverageBV AvBV;
	/**
	  Constructor
		@param t type of the kernel
		@param x intialization of model parameters
		@param y initialization of bias variance values
	*/
	model_BV (int t=0, float x=1, float y=0) : P(x,t), AvBV(y) {}
	/// Prints on the stream out its content
	void print (ostream& out = cout) {P.print(out); AvBV.print(out);}
}; 

/**
   Class building  a lobag ensemble of learning machines.\\
	 
*/ 
class Lobag {

private:


protected:
/// Base learner model. Each new added base learner is equal to base.
BaseSVM_light_common* base;

/// Ensemble of bagged SVM
Bagging * bagSVM;

/// Bootstrapped data used to perform the out-of-bag estimate of the error decomposition
BootstrapLight * b;

/// learning set file
char learning_set [256];

/// number of the data of the dataset
int ndata;

/// number of bootstrap samples
int B;

/// Dimension of the data
int diminput;

/** Kernel type.
0 -> linear\\
1 -> polynomial\\
2 -> gaussian\\
3 -> sigmoid\\
*/
int svm_type;

/// Container of the model parameters with their corresponding loss, bias and variance
deque<model_BV> V;

/// Container of the model parameters to be evaluated
deque<par_model> Param;

/// model of the selected lobag ensemble 
model_BV min_bias;

/// model of the selected bagged ensemble 
model_BV min_loss;

/// minimum bias of the evaluated models
float B_min;

/// minimum loss of the evaluated models
float L_min;

/// minimum bias of the model with the minimum error
float B_L_min;

/// directory where the bootstrapped samples are stored
char * dirbs;

/**
	 Read model parameters.
	 It load a single type of algorithm parameters into a container
	 @param ifs input stream of the parameters to be read
	 @param d  container of the parametrs
*/
void read_param(ifstream &ifs, deque<float>& d);

/**
	 Print model parameters.
	 It prints the parameter of the models specified in mbv.
	 @param mbv model parameters
	 @param out stream where the parameters are printed
*/
void print_oob_parameters(const model_BV& mbv, ostream& out = cout);

/////////////////////////////////////////////////////
public:

/// Default constructor
Lobag();

/**
	 Constructor.
	 Model file format:\\
	 Kernel: k\\
	 c1 c2 c3 ... cn\\
	 end\\
	 d1 d2 d3 ... dn opp\\
	 s1 s2 s3 ... sn \\
	 end\\
	 Feasible values for k are 0 (linear), 1(polynomial), 2(gaussian), 3(sigmoid).\\
	 c1 c2 c3 ... cn correspond to the list of the C regularization values.\\
	 If k=0 the lines strating with d1 and with s1 are not present.\\
	 If k=1 is present only the line d1 d2 d3 ... dn corresponding to the polynomial degree.\\
	 If k=2 or k=3 is present only the line  s1 s2 s3 ... corresponding to the sigma values for gaussian or to the
	 multiplicative coefficient for sigmoid kernels.\\	 
	 @param modelfile file storing the information about the models to be evaluated
	 @param data   file from which data are randomly drawn
	 @param nf number of bootstrapped files to be generated
	 @param seed random seed (def. 0). If seed=0 CPU time is used.
	 @param dir directory where the bootstrapped samples will be stored
*/
Lobag (char* modelfile, char* data, int nf, int seed = 0, char* dir = DIRBS);


/// Destructor
~Lobag();

/**
	 Load model parameters.
	 It load algorithm parameters into a container reading them from a file
	 @param modelfile file of the algorithm parameters
	 @return code of the type of kernel read from modelfile
*/
int load_model_parameters(char* modelfile);


/** It performs the bias-variance decomposition of the error on a specified model.
    The bias variance decomposition of the error is performed on a specific SVM model. The results of the loss, bias
		net-variance, unbiased and biased variance are returned in mbv.
    @param pm model parameters
		@param dir_models directory where the models will be stored
		@param dir_res directory where the results on the overall learning set will be stored
		@param mbv computed bias and variance values
*/
void do_BV_decomposition(par_model pm,  char* dir_models, char* dir_res, model_BV& mbv);

/** Builds a base learner.\\
It builds a base learner according to the model parameters stored in pm.
    @param pm model parameters
		@return the new base learner
*/
BaseSVM_light_common* build_base_learner (par_model pm);


/** Performs in an exhaustive way the bias-variance decomposition of the error.
    The bias variance decomposition of the error is performed on a set of specific SVM model, calling
		iteratively the method do_BV_decomposition. The SVM models are stored in the container Param. 
		The results of the loss, bias net-variance, unbiased and biased variance of each model are stored in the V container. 
		@param dir_models directory where the models will be stored
		@param dir_res directory where the results on the overall learning set will be stored
*/
void exhaustive_BV_decomposition(char* dir_models = DIRMODELS, char* dir_res = DIR_RES);
             
/** 
   Prints the values of bias-variance decomposition for all the models.
	 It prints the values of the loss, bias, net, unbiased and biased variance and of the pure variance
	 of all the considered models.
	 @param out output stream where the values are printed
*/
void print_BV(ostream& out = cout);

/** 
   Save the values of bias-variance decompostion for all the models.
	 It saves the values of the loss, bias, net, unbiased and biased variance and of the pure variance
	 of all the considered models.
	 @param file file where the values will be saved
*/
void save_BV(char* file);

/** 
   Select the model for the lobag ensemble.
	 It selects from the container V the models with the minimum estimated bias (min_bias), and with the minimum
	 estimated loss (min_loss). It evaluates the minimum oob estimate of the error of the lobag ensemble setting
	 the member B_min that corrsponds to the bias of the model with the minimum oob estimate of the bias,
	 and the oob minimum of the error of the bagged ensemble, setting the member B_L_min that corrsponds to 
	 the bias of the model with the minimum oob estimate of the error. The member B_min is set to the oob estimate
	 of the error of the single SVM.
*/
void Select_model(void);

/** 
   Prints the out-of-bag estimate of the error of lobag and the bagged ensemble.\\
	 It prints the oob estimate of the error of the lobag and bagged ensemble and the estimate of the error of the
	 single SVM.
	 @param out output stream where the values are printed
*/
void print_oob_error (ostream& out = cout);

/** 
   Returns the container storing the models and the values of bias--variance decomposition.\\
	 @return container of the models
*/
deque<model_BV> get_V (void) {return V;}

};





#endif
