/*******************************************************************************
matrix.h 
*******************************************************************************/
#ifndef MATRIX_H
#define MATRIX_H

#include<iostream>
#include <fstream>
#include <strstream>
#include <string>
#include <ctype.h>
#include "vect.h"

using namespace std;

/**
   Class for manipulating matrices of real numbers.\\
*/ 
class matrix {

  vect** m;
  int  d1;
  int  d2;
  
	// Check if dimensions of matrices match
  void     check_dimensions(const matrix&) const; 
	
	// return the element (i,j) of the matrix
  double&  el(int i, int j) const { return m[i]->v[j]; }
    
public:

/**
	 Constructor.
	 The matrix is initialized to a m * n  matrix.
	 @param n number of rows
	 @param m number of columns
	 @param val initialization value
*/
  matrix(int n=0, int m=0, double val=0.0);

/**
	 Copy Constructor.
*/
  matrix(const matrix&);

/**
	 Constructor.
	 The matrix is initialized to a vec.dim() * 1  matrix, using the values of the vector vec.
	 @param vec vector
*/  	
  matrix(const vect& vec);

/**
	 Constructor.
	 The matrix is initialized to a n * m  matrix, using the values of the matrix of double pp.
	 @param n number of rows
	 @param m number of columns
	 @param p matrix of double
*/  
	matrix(int n, int m, double** p);

/**
	 Assignment operator.
*/
  matrix& operator=(const matrix&);

/**
	 Destructor.
*/
 ~matrix();

/**
	 Returns the number of rows of the matrix
*/
int     dim1()  const  {  return d1; }

/**
	 Returns the number of columns of the matrix
*/
int     dim2()  const  {  return d2; }


/**
	 Returns the i-th row of the matrix
	 @param i row of the matrix
	 @return vector corresponding to the row i of the matrix 
*/
vect& row(int i) const;

/**
	 Returns the i-th column of the matrix
	 @param i column of the matrix
	 @return vector corresponding to the column i of the matrix 
*/
vect  col(int i) const;

/**
	 Returns the transpose of the matrix
	 @return transpose of the matrix
*/
matrix  transpose() const;


/**
	 Matrix to vect type conversion.
*/
operator vect() const; 

/**
	 Returns a row i of the matrix
	 @param i row index of the matrix
	 @retrun row  of the matrix
*/
vect& operator[](int i)    const { return row(i); }

/**
	 Returns the element (i,j) of the matrix
	 @param i row 
	 @param j column
	 @retrun element (i,j)  of the matrix
*/
double& operator()(int i, int j);

/**
	 Returns the element (i,j) of the matrix
	 @param i row 
	 @param j column
	 @retrun element (i,j)  of the matrix
*/
double  operator()(int,int) const;

/**
	Boolean equality test
*/  
int     operator==(const matrix&)    const;

/**
	Boolean inequality test
*/  
int     operator!=(const matrix& x)  const { return !(*this == x); }

/**
	 Addition of matrices.\\
	 Preconditions: M.dim1() == M1.dim1() and M.dim2() == M1.dim2().
*/ 
matrix operator+(const matrix& M1);

/**
	 Subtraction of matrices.\\
	 Preconditions: M.dim1() == M1.dim1() and M.dim2() == M1.dim2().
*/ 
matrix operator-(const matrix& M1);

/**
	 Unary minus.
*/ 
matrix operator-(); 

/**
	 Overloading -= operator.
*/
matrix& operator-=(const matrix&);

/**
	 Overloading += operator.
*/
matrix& operator+=(const matrix&);

/**
	 Multiplication of matrices.\\
	 Preconditions: M.dim2() == M1.dim1()
*/ 
matrix operator*(const matrix& M1);

/**
	 Multiplication with vector.\\
	 Preconditions: M.dim2() == vec.dim()
*/ 
vect operator*(const vect& vec) { return vect(*this * matrix(vec)); }

/**
	 Multiplication with a scalar.\\
	 Preconditions: M.dim2() == vec.dim()
*/ 
matrix operator*(double x);

/**
	Prints the matrix row by row to ostream O
*/ 
void print(ostream& O);

/**
	Prints the matrix row by row to cout
*/ 
void print() { print(cout); }

/**
	Reads a matrix from the stream I.\\
	Reads M.dim1 times M.dim2() numbers from input stream
  I and writes them row by row into the matrix.
*/ 
void read(istream& I);

/**
	Reads a matrix from standard input.\\
	Reads M.dim1 times M.dim2() numbers from cin
  and writes them row by row into the matrix.
*/ 
void read() { read(cin); }

/**
	Initialize the matrix.\\
	All the elements of the matrix are initialized to value
	@param val initialization value
*/ 
void initialize(double val = 0.0);

/**
	Writes a matrix to an output stream.\\
	Writes the matrix  row by row to the output stream O
*/ 
friend ostream& operator<<(ostream& O, const matrix& M);

/**
	Reads a matrix froam an input stream.\\
	Reads the matrix  row by row from the input stream I
*/ 
friend istream& operator>>(istream& I, matrix& M);

};



#endif
