// mutualinfo7.h
 

#ifndef MUTUAL7_H
#define MUTUAL7_H

#pragma interface
#include<iostream.h>
#include<fstream.h>
#include<iomanip.h>
#include<strstream.h>
#include<string.h>
#include"utils.h"

// Default minimum of the random variable value
  static const float MIN_VALUE = 0.0;
// Default maximum of the random variable value
  static const float MAX_VALUE = 1.0;	


	 
// ============================================================================================

/**
   Class for evaluating mutual information between 7 random variables.\\
   Random variables are discretized dividing their range values in bins. For each bin is
	 computed its frequency, both for each single variable and the joint frequency.
	 Methods can be used to evaluate the entropy of each variable, the joint entropy,
	 the conditional entropy and the mutual information. You can specify a number of
	 2 to 6 equispaced bins for each variable
*/ 


template<int sz> class MutualInfo7 {

protected:
  
	/// Number of random variables
	static const int num_var = 7;
	/// Minimum number of bins
	static const int MIN_BIN = 2;
	///  Maximum number of bins
	static const int MAX_BIN = 6;
	/// Minimum of the random variable value
	float min;
	/// Maximum of the random variable value
	float max;
  /// number of data values of each random variable
	unsigned ndata;
	/// number of bins
  unsigned nbins;
	/// Amplitude of a single bin
	float ampl;
	/// Matrix of values of the the random variables. Num rows = ndata, num columns = num_var
  float **data;
	/// matrix of values of frequencies of random variables; num rows = num_var, num columns = sz 
	float **p;
	/// matrix of joint frequencies between the 7 variables. It is a matrix num_var-dimensional 
	/// with sz elements for each dimension.  
	float Pjoint[sz][sz][sz][sz][sz][sz][sz];
	///  Entropy of the 7 random variables;
	float H[num_var];
	/// Joint entropy of the 7 random variables
	float Hjoint;
	/// Mutual information between the seven variables
	float Minfo;
	/// Pearson's statistics  between the seven variables to test correlation.
	float Pearson;
	

  /// Builds data matrix
  void	builddata(unsigned num_data);
	 
  /**
	 Load random variables data from filef to array data.
	 @param num_data random variable values cardinality
	 @param f file containing the values of random variables. This file must have columns
	 corresponding to different variables separated by at least a blank.
	 @param first the column in the file of the first variable. The columns are numbered from 1 to up.
	 @param last the column in the file of the last variable to be read
	*/ 
  int readdata(unsigned num_data, char *f, unsigned first, unsigned last);
	
	/// It builds and initializes  matrix of marginal frequencies p
  void buildp(unsigned num_bins);
	
	/// It builds and initializes matrix of joint frequencies
  void buildPjoint(unsigned num_bins);

  /**
	 Calculate the bin of random variable num_var od the  data[num_var][num_data].
	 It calaculates the index of the random variable num_var where the value data[num_var][num_data]
	 falls. num_var is the row of p and the column of data. It corresponds also to the dimension of
	 Pjoint.
	 @param num_var index of the random variable in the matrix data
	 @param num_data index of random variable value in data
	 @return index number of the bin of p
	*/ 
  virtual unsigned assignbin(unsigned num_var, unsigned num_data);
	
	/// Initializes p
	void Init_p (float init=0.0);
	
	/// Initializes Pjoint
	void Init_Pjoint (float init=0.0);
	
	/// Controls th number of bins. If it is not in the range, exits.
	void control_bins(void);
	
public:
   
	/**
	 Constructor.
	 @param num_data random variable values cardinality
	 @param thedata matrix of data. It must have 7 columns, each referring to a different random
	 variable
	 @param min_v Minimum of the random variable value
	 @param max_v Maximum of the random variable value
	 variables. 
	*/
  MutualInfo7 (unsigned num_data,  float **thedata,
	             float min_v = 0.0, float max_v = 1.0);
	 
	 
	 /**
		Constructor.
		@param num_data random variable values cardinality
		@param f file containing the values of random variables. This file must have columns
		corresponding to different variables separated by at least a blank.
		@param first the column in the file of the first variable. The columns are numbered from 1 to up
		@param last the column in the file of the last variable
		@param min_v Minimum of the random variable value
		@param max_v Maximum of the random variable value
	 */
	 MutualInfo7 (unsigned num_data,   char *f, 
              	unsigned first, unsigned last, float min_v = 0.0, float max_v = 1.0);

	 
	 /// Destroyer
	 ~MutualInfo7 (); 
	 
	 /// print data to stdout
	 void printdata(void);
	 
	 /// return minimum value of random variable
	 float Read_min() {return min;}
	 /// return maximum value of random variable
	 float Read_max() {return max;}
	 /// print on stdout minimum value of random variable
	 void Print_min() {cout << "Min. of random variable = " << min << endl;}
	 /// print on stdout maximum value of random variable
	 void Print_max() {cout << "Max. of random variable = " << max << endl;}
	 
	 /**
	 It calculates the frequency of the marginal random variables.
	 The random variables are discretized in bins and for each bin is calculated the frequency
	 considering the values of the array data. 
	 */
   void Calc_p (void);
	 
	 /**
	 It calculates the joint frequency of the random variables.
	 The random variables are discretized in bins and for each bin is calculated the frequency
	 considering the values of the array data. 
	 */
	 void  Calc_Pjoint (void);
	 
	 /// print the matrix p
	 void Print_p(void);
	 
	 /// return the sum of the values of Pjoint. Must be 1.
	 float Sum_Pjoint(void);
	 
	 /// Initializes entropy of random variables to init. Init default is 0;
	 void Init_Entropy(float init=0);
	 
	 /**
	 It calculates the entropy of the random variables.
	 @return the pointer to the variables entropy. 
	 */
	 float* Calc_Entropy(void);
	 
	 /// Returns the entropy array of the random variables
	 float* Read_Entropy(void){return H;}
	 
	 /// Print to stdout the entropy of the random variables
	 void Print_Entropy(void);
	 
	 /**
	 It calculates the joint entropy of the random variables.
	 @return joint entropy of the random variables. 
	 */
	 float Calc_JointEntropy(void);
	 
	 /// Returns the joint entropy  of the random variables
	 float Read_JointEntropy(void){return Hjoint;}
	 
	 /// Print to stdout the joint entropy of the random variables
	 void Print_JointEntropy(void);
	 
	 /**
	 It calculates the mutual information of the random variables.
	 @return mutual information of the random variables. 
	 */
	 float Calc_Minfo(void);
	 
	 /// Returns the mutual information  of the random variables
	 float Read_Minfo(void){return Minfo;}
	 
	 /// Print to stdout the mutual information of the random variables
	 void Print_Minfo(void);
	 
	 /**
	 It calculates the Pearson's statistics. It evaluates the correlation between the 
	 7 random variables.
	 @return result of the Pearson's statistics.
	 */
	 float Calc_Pearson(void);
	 
	 /// Print to stdout the Pearson's statistics of the random variables
	 void Print_Pearson(void);
}; 




// ============================================================================================

/**
   Class for evaluating the correlation between 7 random variables representing errors.\\
   Random variables are discretized dividing their range values in bins. For each bin is
	 computed its frequency, both for each single variable and the joint frequency.
	 Methods can be used to evaluate the entropy of each variable, the joint entropy,
	 the conditional entropy and the mutual information. The bins are not necessarily 
	 of the same size. The first bin correspond to correct values, the others to increasing
	 errors. The input data must be errors between the stimated and the "true" values.
	 The Mutual information error is calculated subtracting to the mutual information the mutual
	 information generated by "correct" values or values with a single error.
*/ 


template<int sz> class MIError7 : public  MutualInfo7 <sz> {


protected:

  /// Mutual information error
	float MIErr;
	
	/// Mutual information error minus mutual information generated by a single error
	float MIErr2;
	
	/// Mutual information generated by correct guess or by a single error
	float MICorrect2;
	
	/// Mutual information generated by correct guess
	float MICorrect;
	
	/** array of the values of the limits of the bins. \\
	    The first value correspond to the leftmost value=min, the last value correspond to the
			rightmost=max. The size of the vector must be nbins + 1.
	*/
	float * bin; 
	
	/** It builds the values of the limits of the bins. \\
	    The first value correspond to the leftmost value=min, the last value correspond to the
			rightmost=max. If binlimits == 0 are builts equispaced bins.
			@param binlimits. Array of the values of the bin limits. The size of the vector must 
			be nbins + 1.
			@see bin
	*/
  void buildbin(float *binlimits = 0);
	
	/**
		Calculate the bin of random variable num_var of the  data[num_data][num_var].
		It calculates the index of the random variable num_var where the value data[num_data][num_var]
		falls. num_var is the row of p and the column of data. It corresponds also to the dimension of
		Pjoint.
		@param num_var index of the random variable in the matrix data
		@param num_data index of random variable value in data
		@return index number of the bin of p
	*/ 
  virtual unsigned assignbin(unsigned num_var, unsigned num_data);
	
public:

  /**
		Constructor.
		@param num_data random variable values cardinality
		@param thedata matrix of data. It must have 7 columns, each referring to a different random
		variable
		@param binlimits Array of the limit values of the bins. Its size must be sz + 1. If it is
		not specified the bins are equally spaced between min_v and max_v
		@param min_v Minimum of the random variable value
		@param max_v Maximum of the random variable value
		variables. 
	*/
  MIError7 (unsigned num_data,  float **thedata, float *binlimits = 0,
	             float min_v = 0.0, float max_v = 1.0);
							 
	/**
	 Constructor.
		@param num_data random variable values cardinality
		@param f file containing the values of random variables. This file must have columns
		corresponding to different variables separated by at least a blank.
		@param first the column in the file of the first variable. The columns are numbered from1 to up
		@param last the column in the file of the last variable
		@param binlimits Array of the limit values of the bins. Its size must be sz + 1. If it is
		not specified the bins are equally spaced between min_v and max_v
		@param min_v Minimum of the random variable value
		@param max_v Maximum of the random variable value
	*/
	MIError7 (unsigned num_data,   char *f, unsigned first, unsigned last, 
            float *binlimits = 0, float min_v = 0.0, float max_v = 1.0);
  
	/// Destructor
  ~MIError7 (); 
	
	/** It calculates the component of mutual information generated by no errors or not correlated
	     errors. \\
			 Considering the matrix of the element of the mutual information between the 7 discretized
			 variables, it is computed the mutual info corresponding to the 7 "edges" departing from
			 the vertex corresponding to the correct bin (i.e. the vertex with bin[0] on all the 7
			 dimension of the hypercube).
	     @return the component of mutual information generated by no errors or not correlated
	     errors
	*/
	float Calc_MICorrect2(void);
	
	/** It calculates the component of mutual information generated by no errors.\\
			 Considering the matrix of the element of the mutual information between the 7 discretized
			 variables, it is computed the mutual info corresponding to the 
			 vertex of the correct bin (i.e. the vertex with bin[0] on all the 7
			 dimension of the hypercube).
	     @return the component of mutual information generated by no errors or not correlated
	     errors
	*/
	float Calc_MICorrect(void);
	 
	 
	/** It calculates the component of mutual information generated by correlated errors. \\
			 Considering the matrix of the element of the mutual information between the 7 discretized
			 variables, is computed the mutual info corresponding to the overall matrix
			 minus the 7 "edges" departing from the vertex corresponding to the correct bin.
	     @return the component of mutual information generated by correlated errors.
	*/
  float Calc_MIErr2(void); 
	
	/** It calculates the component of mutual information generated by  errors. \\
			 Considering the matrix of the element of the mutual information between the 7 discretized
			 variables, is computed the mutual info corresponding to the overall matrix
			 minus the  vertex corresponding to the correct bin. It considers all errors, also single.
	     @return the component of mutual information generated by all the errors.
	*/
  float Calc_MIErr(void); 
	
	
  /** Print to stdout the mutual information MICorrect2. \\
			Print to stdout the mutual information MICorrect2 generated by no errors 
			or not correlated errors
	*/
  void Print_MICorrect2(void);	

  /** Print to stdout the mutual information MIErr2. \\
			Print to stdout the mutual information generated by correlated errors.
	*/
  void Print_MIErr2(void);	
	
	/** Print to stdout the mutual information MICorrect. \\
			Print to stdout the mutual information MICorrect generated by no errors 
	*/
  void Print_MICorrect(void);	

  /** Print to stdout the mutual information MIErr. \\
			Print to stdout the mutual information generated by errors.
	*/
  void Print_MIErr(void);
	
	/// It reads MiErr2
	float Read_MiErr2(void) {return MIErr2;}	
	
	/// It reads MiErr
	float Read_MiErr(void) {return MIErr;}	
  
}; // end of class MIError7









#endif

	 
