// roc.h


#ifndef ROC_H
#define ROC_H

#include<iostream>
#include<fstream>
#include<vector>
#include<string>

using namespace std;

/**
   Class storing quantities used for ROC.\\
   This class stores values of sensitivity, specificity, 1-specificity and precision, used to compute ROC
	 curves.
*/ 
class ROCvalue {
  /// It measures the proportion of correctly predicted among all positive examples.
	/// In literature it is also named recall.
  float sensitivity;
	/// It measures the proportion of correctly predicted among all negative examples
	float specificity;
	/// It measures the proportion of incorrectly predicted among all negative examples
	float comp_specificity;
	/// It measures the proportion of correctly predicted among all predicted positive examples
	float precision;
	
	public:
	friend class ROC;
	/// Constructor
	ROCvalue () {
	  sensitivity = specificity = comp_specificity = precision = 0;}
	/// It returns the sensitivity
	float get_sensitivity() const {return sensitivity;}
	/// It returns the specificity
	float get_specificity() const {return specificity;}
	/// It returns 1 - specificity
	float get_comp_specificity() const {return comp_specificity;}
	/// It returns the precision
	float get_precision() const {return precision;}
	/// It prints on stdout the ROCvalues pairs 1 - specificity and sensitivity
	void print_ROC() const;
	/// It prints on stdout all the ROCvalues
	void print_all() const;
	/// It prints on stdout the cardinality of the data
	void print_numdata() const;
};



/**
   It computes the ROC curves.\\
   This class computes Receiver Operating Characteristic (ROC) curves for learning machines having a threshold that separates positive from
	 negative patterns. It computes sensitivty, specificity and precision for different threshold values.
*/ 
class ROC {
  /// Dimension of confusion matrix
  static const int size_matrix = 2;
	/// Vector of the predicted values
  vector <float> predicted;
	/// Vector of the target values
	vector <float> target;
	/// Cardinality of the data
	int numdata;
	/// Vector storing the points for the ROC curve
	vector <ROCvalue> rocvalues;
	
	/// Threshold value to separate positive from negative patterns
	float threshold;
	/// Confusion matrix
	int conf_matrix [size_matrix][size_matrix];
	/// initialize confusion matrix to 0
	void init_conf_matrix(void);
	/// It computes the minimum of the predicted values
	float calc_min(void);
	/// It computes the maximum of the predicted values
	float calc_max(void);
	
	public:
	/// Default constructor
	ROC(){
	  init_conf_matrix();
		threshold = 0;
		numdata = 0;
	}
	
	/**
	 Constructor. \\
	 It reads from a file the predicted and the target values.
	 The first column of the data specifies the predicted values and the second the targets.
	 File format: one predicted value and one target value for each line separated by blanks.
	 The predicted values are loaded in the vector predicted and the targets in the vector target.
	 @param filename name of the data file
	*/ 
	ROC(char * filename){
	  init_conf_matrix();
		threshold = 0;
		numdata = 0;
		read_data(filename);
	}
	
	/**
	  Constructor. \\
	 It reads from two different files the predicted and the target values.
	 The first file stores the predicted values and the second the targets.
	 The predicted values are loaded in the vector predicted and the targets in the vector target.
	 @param pred_file name of the  file storing the predicted values
	 @param target_file name of the  file storing the target values
	*/ 
	ROC(char * pred_file, char * target_file){
	  init_conf_matrix();
		threshold = 0;
		numdata = 0;
		read_data_2(pred_file, target_file);
	}
	
	/**
	 Load predicted and target data from a file. \\
	 The first column of the data specifies the predicted values and the second the targets.
	 File format: one predicted value and one target value for each line separated by blanks.
	 The predicted values are loaded in the vector predicted and the targets in the vector target.
	 @param filename name of the data file
	*/ 
	void read_data(char * filename);
	
	/**   
	 Load predicted and target data from two separated files. \\
	 The first file stores the predicted values and the second the targets.
	 The predicted values are loaded in the vector predicted and the targets in the vector target.
	 @param pred_file name of the  file storing the predicted values
	 @param target_file name of the  file storing the target values
	*/ 
	void read_data_2(char * pred_file, char * target_file);
	
	/**
	 It builds a confusion matrix. \\
	 It reads the predicted and target vectors in order to calculate the confusion matrix using the
	 threshold thresh. If the predicted value is above thresh it is considered ad a positive response,
	 otherwise it is considered as negative.
	 @param thresh Threshold used to compute the confusion matrix
	*/ 
	void do_conf_matrix(float thresh=0);

  /**
	 It computes a ROCvalue. \\
	 It calculates the ROCvalues (sensitivity, specificity and precision), using a threshold to
	 separate positive from negative examples.
	 @param thresh Threshold used to compute the confusion matrix
	 @return a pointer to the computed ROC value
	*/ 
	ROCvalue * compute_ROC_value (float thresh); 
	
	/**
	 It computes a ROC curve \\
	 It computes a ROC curve, subdividing the output in a set of equidistant intervals. The limits of
	 the intervals correspond to a different threshold and to a different computed ROC value.
	 @param num_interval number of intervals
	*/ 
	void compute_ROC_curve (int num_interval); 
	
	/**
	 It prints a ROC curve \\
	 It prints a ROC curve on stdout, printing sensitivity and 1-specificity pairs.
	*/ 
	void print_ROC_curve (void) const; 
	
	/**
	 It prints a ROC curve considering also precision values \\
	 It prints a ROC curve on stdout, printing sensitivity, specificity, 1-specificity and precision.
	*/ 
	void print_ROC_curve_all (void) const; 
	
	///It prints the predicted data on stdout. 
	void print_predicted (void) const; 
	
	///It prints the target data on stdout. 
	void print_target (void) const; 
};


#endif
